# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pylint: disable=too-many-public-methods
# pyright: reportDeprecated=false,reportUnknownMemberType=false
# pyright: reportAttributeAccessIssue=false,reportUnknownVariableType=false

"""Waldiez chat model."""

from typing import TYPE_CHECKING, Any, Optional

from pydantic import Field, model_validator
from typing_extensions import Annotated, Literal, Self

from waldiez.models.chat.chat_summary import WaldiezChatSummary

from ..common import (
    WaldiezAgentTarget,
    WaldiezBase,
    WaldiezHandoff,
    WaldiezHandoffCondition,
    WaldiezTransitionAvailability,
    WaldiezTransitionTarget,
    generate_function,
)
from .chat_data import WaldiezChatData
from .chat_message import (
    CALLABLE_MESSAGE,
    CALLABLE_MESSAGE_ARGS,
    CALLABLE_MESSAGE_RAG_WITH_CARRYOVER_TYPES,
    CALLABLE_MESSAGE_TYPES,
    RAG_METHOD_WITH_CARRYOVER_BODY,
    WaldiezChatMessage,
)
from .chat_nested import (
    NESTED_CHAT_ARGS,
    NESTED_CHAT_MESSAGE,
    NESTED_CHAT_REPLY,
    NESTED_CHAT_TYPES,
    WaldiezChatNested,
)
from .chat_summary import (
    CALLABLE_SUMMARY,
    CALLABLE_SUMMARY_ARGS,
    CALLABLE_SUMMARY_TYPES,
)

if TYPE_CHECKING:
    from ..agents import WaldiezAgent

WaldiezChatType = Literal["chat", "nested", "group", "hidden"]
"""Possible chat types: "chat", "nested", "group", "hidden"."""


# noinspection PyUnresolvedReferences
class WaldiezChat(WaldiezBase):
    """Chat class.

    Attributes
    ----------
    id : str
        The chat ID.
    source : str
        The source of the chat (sender).
    target : str
        The target of the chat (recipient).
    type : WaldiezChatType
        The type of the chat data: "chat", "nested", "group", or "hidden".
    data : WaldiezChatData
        The chat data.
        See `waldiez.models.chat.WaldiezChatData` for more information.
    name : str
        The chat name.
    nested_chat : WaldiezChatNested
        The nested chat message/reply if any.
    message : WaldiezChatMessage
        The chat message.
    message_content : str | None
        The chat message content if any. If method, the method's body.

    Functions
    ---------
    get_chat_args()
        Get the chat arguments to use in autogen.
    """

    id: Annotated[
        str,
        Field(
            ...,
            title="ID",
            description="The chat ID.",
        ),
    ]
    type: Annotated[
        WaldiezChatType,
        Field(
            default="chat",
            title="Type",
            description="The type of the chat data.",
        ),
    ]
    source: Annotated[
        str,
        Field(
            ...,
            title="Source",
            description="The chat source.",
        ),
    ]
    target: Annotated[
        str,
        Field(
            ...,
            title="Target",
            description="The chat target.",
        ),
    ]
    data: Annotated[
        WaldiezChatData,
        Field(
            ...,
            title="Data",
            description="The chat data.",
        ),
    ]

    @property
    def name(self) -> str:
        """Get the name."""
        return self.data.name

    @property
    def description(self) -> str:
        """Get the description."""
        return self.data.description

    @property
    def order(self) -> int:
        """Get the order."""
        return self.data.order

    @property
    def nested_chat(self) -> WaldiezChatNested:
        """Get the nested chat."""
        return self.data.nested_chat

    @property
    def message(self) -> WaldiezChatMessage:
        """Get the message."""
        if isinstance(
            self.data.message, str
        ):  # pragma: no cover (just for the lint)
            return WaldiezChatMessage(
                type="string",
                use_carryover=False,
                content=self.data.message,
                context={},
            )
        return self.data.message

    @property
    def summary(self) -> WaldiezChatSummary:
        """Get the summary."""
        return self.data.summary

    @property
    def message_content(self) -> str | None:
        """Get the message content."""
        return self.data.message_content

    @property
    def summary_content(self) -> str | None:
        """Get the summary content."""
        return self.data.summary_content

    @property
    def max_turns(self) -> int | None:
        """Get the max rounds for the chat."""
        return self.data.max_turns

    @property
    def chat_id(self) -> int:
        """Get the chat ID."""
        return self.data.get_chat_id()

    @property
    def prerequisites(self) -> list[int]:
        """Get the chat prerequisites."""
        return self.data.get_prerequisites()

    @property
    def condition(self) -> WaldiezHandoffCondition:
        """Get the handoff condition."""
        return self.data.condition

    @property
    def available(self) -> WaldiezTransitionAvailability:
        """Get the transition availability."""
        return self.data.available

    @model_validator(mode="after")
    def _validate_chat(self) -> Self:
        """Override if needed the source and target of the chat."""
        if self.data.real_source:
            self.source = self.data.real_source
        if self.data.real_target:
            self.target = self.data.real_target
        return self

    def as_handoff(
        self,
    ) -> WaldiezHandoff:
        """Convert the chat to a handoff.

        Returns
        -------
        WaldiezHandoff
            The handoff representation of the chat.
        """
        # noinspection PyTypeHints
        target: WaldiezTransitionTarget = WaldiezAgentTarget(
            target_type="AgentTarget",
            value=[self.target],
        )
        return WaldiezHandoff(
            target=target,
            condition=self.condition,
            available=self.available,
        )

    def set_chat_id(self, value: int) -> None:
        """Set the chat ID.

        Parameters
        ----------
        value : int
            The chat ID.
        """
        self.data.set_chat_id(value)

    def set_prerequisites(self, value: list[int]) -> None:
        """Set the chat prerequisites.

        Parameters
        ----------
        value : list[int]
            The chat prerequisites.
        """
        self.data.set_prerequisites(value)

    def get_message_function(
        self,
        name_prefix: str | None = None,
        name_suffix: str | None = None,
        is_rag: bool = False,
    ) -> tuple[str, str]:
        """Get the message function.

        Parameters
        ----------
        name_prefix : str
            The function name prefix.
        name_suffix : str
            The function name suffix.
        is_rag : bool
            If the message is from a RAG user.

        Returns
        -------
        tuple[str, str]
            The message function and the function name.
        """
        if self.message.type in ("string", "none") or (
            not self.message_content and is_rag is False
        ):
            return "", ""
        function_types = CALLABLE_MESSAGE_TYPES
        function_name = CALLABLE_MESSAGE
        if name_prefix:
            function_name = f"{name_prefix}_{function_name}"
        if name_suffix:
            function_name = f"{function_name}_{name_suffix}"
        if is_rag and self.message.type == "rag_message_generator":
            function_types = CALLABLE_MESSAGE_RAG_WITH_CARRYOVER_TYPES
            return (
                generate_function(
                    function_name=function_name,
                    function_args=CALLABLE_MESSAGE_ARGS,
                    function_types=function_types,
                    function_body=self.message.content_body
                    or RAG_METHOD_WITH_CARRYOVER_BODY,
                ),
                function_name,
            )
        return (
            generate_function(
                function_name=function_name,
                function_args=CALLABLE_MESSAGE_ARGS,
                function_types=function_types,
                function_body=self.message_content or "",
            ),
            function_name,
        )

    def get_summary_function(
        self,
        name_prefix: str | None = None,
        name_suffix: str | None = None,
    ) -> tuple[str, str]:
        """Get the summary function.

        Parameters
        ----------
        name_prefix : str
            The function name prefix.
        name_suffix : str
            The function name suffix.

        Returns
        -------
        tuple[str, str]
            The message function and the function name.
        """
        if self.summary.method != "custom":
            return "", ""
        function_types = CALLABLE_SUMMARY_TYPES
        function_name = CALLABLE_SUMMARY
        if name_prefix:
            function_name = f"{name_prefix}_{function_name}"
        if name_suffix:
            function_name = f"{function_name}_{name_suffix}"
        return (
            generate_function(
                function_name=function_name,
                function_args=CALLABLE_SUMMARY_ARGS,
                function_types=function_types,
                function_body=self.summary_content or "",
            ),
            function_name,
        )

    @staticmethod
    def _get_nested_chat_function(
        content: str | None,
        function_name_base: str,
        name_prefix: str | None = None,
        name_suffix: str | None = None,
    ) -> tuple[str, str]:
        """Get the nested chat function.

        Parameters
        ----------
        content : str
            The content for the function body.
        function_name_base : str
            The base name of the function.
        name_prefix : str
            The function name prefix.
        name_suffix : str
            The function name suffix.

        Returns
        -------
        tuple[str, str]
            The generated function and its name.
        """
        if not content:
            return "", ""
        function_name = function_name_base
        if name_prefix:
            function_name = f"{name_prefix}_{function_name}"
        if name_suffix:
            function_name = f"{function_name}_{name_suffix}"
        return (
            generate_function(
                function_name=function_name,
                function_args=NESTED_CHAT_ARGS,
                function_types=NESTED_CHAT_TYPES,
                function_body=content,
            ),
            function_name,
        )

    def get_nested_chat_message_function(
        self,
        name_prefix: str | None = None,
        name_suffix: str | None = None,
    ) -> tuple[str, str]:
        """Get the nested chat message function.

        Parameters
        ----------
        name_prefix : str | None
            The function name prefix.
        name_suffix : str | None
            The function name suffix.

        Returns
        -------
        tuple[str, str]
            The nested chat message function and the function name.
        """
        if not self.nested_chat.message or (
            self.nested_chat.message.use_carryover is False
            and self.nested_chat.message.type
            in (
                "string",
                "none",
            )
        ):
            return "", ""
        return self._get_nested_chat_function(
            content=self.nested_chat.message_content,
            function_name_base=NESTED_CHAT_MESSAGE,
            name_prefix=name_prefix,
            name_suffix=name_suffix,
        )

    def get_nested_chat_reply_function(
        self,
        name_prefix: str | None = None,
        name_suffix: str | None = None,
    ) -> tuple[str, str]:
        """Get the nested chat reply function.

        Parameters
        ----------
        name_prefix : str | None
            The function name prefix.
        name_suffix : str | None
            The function name suffix.

        Returns
        -------
        tuple[str, str]
            The nested chat reply function and the function name.
        """
        if not self.nested_chat.reply or (
            self.nested_chat.reply.use_carryover is False
            and self.nested_chat.reply.type
            in (
                "string",
                "none",
            )
        ):
            return "", ""
        return self._get_nested_chat_function(
            content=self.nested_chat.reply_content,
            function_name_base=NESTED_CHAT_REPLY,
            name_prefix=name_prefix,
            name_suffix=name_suffix,
        )

    def get_chat_args(
        self,
        for_queue: bool,
        sender: Optional["WaldiezAgent"] = None,
    ) -> dict[str, Any]:
        """Get the chat arguments to use in autogen.

        Parameters
        ----------
        for_queue : bool
            Whether to get the chat arguments for a chat queue.
        sender : WaldiezAgent, optional
            The sender agent, to check if it's a RAG user.

        Returns
        -------
        dict
            The chat arguments.
        """
        args_dict = self.data.get_chat_args(for_queue)
        if (
            sender
            and sender.is_rag_user
            and self.message.type == "rag_message_generator"
            and hasattr(sender.data, "retrieve_config")
        ):
            # check for n_results in agent data, to add in context
            n_results = sender.data.retrieve_config.n_results
            if isinstance(n_results, int) and n_results > 0:
                args_dict["n_results"] = n_results
        return args_dict

    def model_dump(self, **kwargs: Any) -> dict[str, Any]:
        """Dump the model to a dict including the chat attributes.

        Parameters
        ----------
        kwargs : Any
            The keyword arguments.

        Returns
        -------
        dict[str, Any]
            The model dump with the chat attributes.
        """
        dump = super().model_dump(**kwargs)
        dump["name"] = self.name
        dump["description"] = self.description
        dump["nested_chat"] = self.nested_chat.model_dump()
        dump["message"] = self.message.model_dump()
        dump["message_content"] = self.message_content
        dump["max_turns"] = self.max_turns
        dump["summary"] = self.summary.model_dump()
        dump["summary_content"] = self.summary_content
        return dump



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=import-error,line-too-long
# pyright: reportUnknownMemberType=false,reportUnknownVariableType=false
# pyright: reportUnknownArgumentType=false,reportAttributeAccessIssue=false
# pyright: reportGeneralTypeIssues=false,reportAny=false
# flake8: noqa: E501
"""Utilities for WebSocket server management."""

import asyncio
import json
import logging
import socket
import time
from contextlib import closing
from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING, Any, final

try:
    import websockets  # type: ignore[unused-ignore, unused-import, import-not-found, import-untyped] # noqa
except ImportError:  # pragma: no cover
    from ._mock import websockets  # type: ignore[no-redef,unused-ignore]


if TYPE_CHECKING:
    from .server import WaldiezWsServer


@dataclass
class ErrorStats:
    """Error stats."""

    total_errors: int
    error_counts: dict[str, int]
    most_common_error: str | None = None

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary.

        Returns
        -------
        dict[str, Any]
            Dictionary representation of the error stats
        """
        return asdict(self)


@dataclass
class ServerHealth:
    """Server health status."""

    status: str  # "healthy", "degraded", "unhealthy"
    uptime_seconds: float
    active_connections: int
    total_connections: int
    messages_received: int
    messages_sent: int
    memory_usage_mb: float | None = None
    timestamp: float = 0.0
    error_stats: ErrorStats | None = None

    def __post_init__(self) -> None:
        """Set timestamp after initialization."""
        if self.timestamp == 0.0:
            self.timestamp = time.time()
        if not self.error_stats:
            self.error_stats = ErrorStats(
                total_errors=0, error_counts={}, most_common_error=None
            )

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary.

        Returns
        -------
        dict[str, Any]
            Dictionary representation of the health status
        """
        my_dict = asdict(self)
        my_dict["error_stats"] = (
            self.error_stats.to_dict() if self.error_stats else None
        )
        return my_dict


@final
class HealthChecker:
    """Health checker for WebSocket server."""

    def __init__(self, server: "WaldiezWsServer"):
        """Initialize health checker.

        Parameters
        ----------
        server : WaldiezWsServer
            Server instance to monitor
        """
        self.server = server
        self.check_interval = 30.0  # seconds
        self.task: asyncio.Task[Any] | None = None
        self.last_health: ServerHealth | None = None

    def start(self) -> None:
        """Start health monitoring."""
        if self.task and not self.task.done():
            return

        self.task = asyncio.create_task(self._monitor_loop())

    def stop(self) -> None:
        """Stop health monitoring."""
        if self.task and not self.task.done():  # pragma: no branch
            self.task.cancel()

    async def _monitor_loop(self) -> None:
        """Health monitoring loop."""
        while True:
            # pylint: disable=too-many-try-statements,broad-exception-caught
            try:
                await asyncio.sleep(self.check_interval)
                health = await self.check_health()
                self.last_health = health

                # Log health status
                if health.status != "healthy":
                    logger = logging.getLogger(__name__)
                    logger.warning("Server health: %s", health.status)

            except asyncio.CancelledError:
                break
            except Exception as e:
                logger = logging.getLogger(__name__)
                logger.error("Health check error: %s", e)

    async def check_health(self) -> ServerHealth:
        """Check server health.

        Returns
        -------
        ServerHealth
            Current health status
        """
        stats = self.server.get_stats()
        # error_handler = self.server.error_handler

        # Determine health status
        status = "healthy"

        # Check for error rate
        total_messages = stats["messages_received"] + stats["messages_sent"]
        if total_messages > 0:
            error_rate = stats["errors_total"] / total_messages
            if error_rate > 0.1:  # More than 10% errors
                status = "unhealthy"
            elif error_rate > 0.05:  # More than 5% errors
                status = "degraded"

        # Check if server is running
        if not stats["is_running"]:
            status = "unhealthy"

        # Get memory usage (optional)
        memory_usage: float | None = None
        try:
            # pylint: disable=import-outside-toplevel
            # noinspection PyUnusedImports
            import psutil

            process = psutil.Process()
            memory_usage = process.memory_info().rss / 1024 / 1024  # MB
        except ImportError:
            pass

        return ServerHealth(
            status=status,
            uptime_seconds=stats["uptime_seconds"],
            active_connections=stats["connections_active"],
            total_connections=stats["connections_total"],
            messages_received=stats["messages_received"],
            messages_sent=stats["messages_sent"],
            memory_usage_mb=memory_usage,
            error_stats=stats["error_stats"],
        )

    def get_last_health(self) -> ServerHealth | None:
        """Get last health check result.

        Returns
        -------
        Optional[ServerHealth]
            Last health status or None if no check performed
        """
        return self.last_health


# noinspection PyBroadException
@final
class ConnectionManager:
    """Manages WebSocket connections and provides utilities."""

    def __init__(self, server: "WaldiezWsServer"):
        """Initialize connection manager.

        Parameters
        ----------
        server : WaldiezWsServer
            Server instance
        """
        self.server = server

    async def ping_all_clients(self) -> dict[str, bool]:
        """Ping all connected clients.

        Returns
        -------
        dict[str, bool]
            Map of client_id to ping success status
        """
        results: dict[str, bool] = {}
        ping_message: dict[str, Any] = {
            "type": "pong",
            "timestamp": time.time(),
        }

        for client_id, client in self.server.clients.items():
            try:
                success = await client.send_message(ping_message)
                results[client_id] = success
            except Exception:  # pylint: disable=broad-exception-caught
                results[client_id] = False

        return results

    async def disconnect_client(
        self, client_id: str, reason: str = "Server initiated"
    ) -> bool:
        """Disconnect a specific client.

        Parameters
        ----------
        client_id : str
            Client to disconnect
        reason : str
            Disconnection reason

        Returns
        -------
        bool
            True if client was disconnected successfully
        """
        if client_id not in self.server.clients:
            return False

        client = self.server.clients[client_id]
        try:
            await client.send_message(
                {
                    "type": "disconnect",
                    "reason": reason,
                    "timestamp": time.time(),
                }
            )
            await client.websocket.close(code=1000, reason=reason)
            return True
        except Exception:  # pylint: disable=broad-exception-caught
            return False

    def get_client_info(self, client_id: str) -> dict[str, Any] | None:
        """Get information about a specific client.

        Parameters
        ----------
        client_id : str
            Client ID

        Returns
        -------
        dict[str, Any] | None
            Client information or None if not found
        """
        if client_id not in self.server.clients:
            return None

        client = self.server.clients[client_id]
        return {
            "client_id": client_id,
            "remote_address": client.remote_address,
            "user_agent": client.user_agent,
            "is_active": client.is_active,
            "connection_time": client.connection_time,
            "connection_duration": client.connection_duration,
        }

    def list_clients(self) -> dict[str, dict[str, Any]]:
        """List all connected clients.

        Returns
        -------
        dict[str, dict[str, Any]]
            Map of client_id to client information
        """
        clients = {
            client_id: self.get_client_info(client_id)
            for client_id in self.server.clients
        }
        return {
            client_id: info
            for client_id, info in clients.items()
            if info is not None
        }


async def test_server_connection(
    host: str = "localhost", port: int = 8765, timeout: float = 5.0
) -> dict[str, Any]:
    """Test connection to WebSocket server.

    Parameters
    ----------
    host : str
        Server host
    port : int
        Server port
    timeout : float
        Connection timeout

    Returns
    -------
    dict[str, Any]
        Test results
    """
    start_time = time.time()
    result: dict[str, Any] = {
        "success": False,
        "error": None,
        "response_time_ms": 0.0,
        "server_response": None,
    }
    # pylint: disable=too-many-try-statements,broad-exception-caught
    try:
        uri = f"ws://{host}:{port}"

        async with websockets.connect(
            uri,
            ping_interval=None,
        ) as websocket:
            # Send ping message
            ping_msg = {"action": "ping"}
            await websocket.send(json.dumps(ping_msg))

            # Wait for response
            response = await asyncio.wait_for(websocket.recv(), timeout=timeout)
            result["server_response"] = json.loads(response)
            result["success"] = True

    except asyncio.TimeoutError:
        result["error"] = "Connection timeout"
    except ConnectionRefusedError:
        result["error"] = "Connection refused - server may not be running"
    except Exception as e:
        result["error"] = str(e)

    result["response_time_ms"] = (time.time() - start_time) * 1000
    return result


# noinspection PyBroadException
def is_port_available(port: int) -> bool:
    """Check if the port is available.

    Parameters
    ----------
    port : int
        Port number

    Returns
    -------
    bool
        True if port is available
    """
    # Check IPv4
    try:
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
            sock.bind(("", port))
    except BaseException:  # pylint: disable=broad-exception-caught
        return False

    # Check IPv6
    try:  # pragma: no cover
        with closing(
            socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
        ) as sock:
            # Disable dual-stack to only check IPv6
            sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 1)
            sock.bind(("", port))
    except BaseException:  # pylint: disable=broad-exception-caught
        return False

    return True


def get_available_port() -> int:
    """Get an available port.

    Returns
    -------
    int
        An available port number
    """
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as soc:
        soc.bind(("", 0))
        soc.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return soc.getsockname()[1]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Exporter core exceptions module."""


class ExporterError(Exception):
    """Base exception for exporter errors."""


class ExporterInitializationError(ExporterError):
    """Exception raised when exporter initialization fails."""


class ExporterValidationError(ExporterError):
    """Exception raised when export validation fails."""


class ExporterContentError(ExporterError):
    """Exception raised when content generation fails."""



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""RAG user proxy agent configuration processor."""

from waldiez.models import (
    WaldiezAgent,
    WaldiezRagUserProxy,
    WaldiezRagUserProxyModels,
    WaldiezRagUserProxyRetrieveConfig,
)

from ...core import (
    DefaultPathResolver,
    DefaultSerializer,
    ImportPosition,
    ImportStatement,
    InstanceArgument,
    PathResolver,
    Serializer,
)
from ...core.extras.agent_extras import RAGUserExtras
from .rag import VectorDBExtras, get_vector_db_extras


# pylint: disable=too-few-public-methods
class RagUserProxyAgentProcessor:
    """Processor for RAG user proxy agent configuration."""

    _before_agent: str = ""

    def __init__(
        self,
        agent: WaldiezAgent,
        agent_name: str,
        model_names: dict[str, str],
        path_resolver: PathResolver | None = None,
        serializer: Serializer | None = None,
    ):
        """Initialize the processor.

        Parameters
        ----------
        agent : WaldiezAgent
            The Waldiez RAG user proxy agent to process.
        model_names : dict[str, str]
            A mapping from model id to model name.
        path_resolver : PathResolver | None
            Optional path resolver for resolving paths.
            Defaults to DefaultPathResolver if not provided.
        serializer : Serializer | None
            Optional serializer for the RAG configuration.
            Defaults to DefaultSerializer if not provided.
        """
        self.agent = agent
        self.agent_name = agent_name
        self.model_names = model_names
        self.path_resolver = path_resolver or DefaultPathResolver()
        self.serializer = serializer or DefaultSerializer()

    def process(self) -> RAGUserExtras:
        """Process RAG user proxy agent configuration.

        Returns
        -------
        RAGConfig
            The processed result containing extra arguments, before content,
            imports, and environment variables.
        """
        result = RAGUserExtras(self.agent.id)
        if not self.agent.is_rag_user or not isinstance(
            self.agent, WaldiezRagUserProxy
        ):
            return result
        # Get the extra args
        vector_db_extras = get_vector_db_extras(
            agent=self.agent, agent_name=self.agent_name
        )
        before_agent, arg_value = self._get_retrieve_config(vector_db_extras)
        if before_agent:
            self._before_agent += before_agent
        if arg_value:
            retrieve_arg = InstanceArgument(
                instance_id=self.agent.id,
                name="retrieve_config",
                value=arg_value,
                tabs=1,
            )
            result.add_arg(retrieve_arg)
        for import_statement in vector_db_extras.imports:
            result.add_import(
                ImportStatement(
                    statement=import_statement,
                    position=ImportPosition.THIRD_PARTY,
                )
            )
        if self._before_agent:
            result.prepend_before_agent(self._before_agent)
        return result

    def _get_retrieve_config(
        self, vector_db_extras: VectorDBExtras
    ) -> tuple[str, str]:
        """Get the retrieve config argument for the agent.

        Returns
        -------
        InstanceArgument | None
            The retrieve config argument if applicable, otherwise None.
        """
        if not isinstance(self.agent, WaldiezRagUserProxy):
            return "", ""
        retrieve_config: WaldiezRagUserProxyRetrieveConfig = (
            self.agent.retrieve_config
        )
        if not retrieve_config:
            return "", ""  # pyright: ignore[reportUnreachable]
        args_dict = self._get_args_dict()
        if not args_dict:
            return "", ""
        before_agent = vector_db_extras.before_arg
        if self.agent.retrieve_config.use_custom_token_count:
            function_content, token_count_arg_name = (
                retrieve_config.get_custom_token_count_function(
                    name_suffix=self.agent_name
                )
            )
            args_dict["custom_token_count_function"] = token_count_arg_name
            before_agent += "\n" + function_content + "\n"
        if self.agent.retrieve_config.use_custom_text_split:
            function_content, text_split_arg_name = (
                retrieve_config.get_custom_text_split_function(
                    name_suffix=self.agent_name
                )
            )
            args_dict["custom_text_split_function"] = text_split_arg_name
            before_agent += "\n" + function_content + "\n"
        if before_agent.strip() and not before_agent.endswith("\n"):
            before_agent += "\n"
        return before_agent, self._get_args_string(
            args_dict,
            vector_db_extras,
            before_agent,
        )

    def _get_args_string(
        self,
        args_dict: dict[str, str | list[str]],
        vector_db_extras: VectorDBExtras,
        before_agent: str,
    ) -> str:
        if not isinstance(self.agent, WaldiezRagUserProxy):
            return ""
        args_content = self.serializer.serialize(args_dict)
        # get the last line (where the dict ends)
        args_parts = args_content.split("\n")
        before_vector_db = args_parts[:-1]
        closing_arg = args_parts[-1]
        args_content = "\n".join(before_vector_db)
        # add the vector_db arg
        args_content += (
            ",\n"
            + f'        "vector_db": {vector_db_extras.vector_db_arg},'
            + "\n"
        )
        # we should not need to include the client, but let's do it
        # to avoid later issues (with telemetry or other client settings)
        # https://github.com/ag2ai/ag2/blob/main/autogen/agentchat/\
        #   contrib/retrieve_user_proxy_agent.py#L265-L266
        if (
            f"{self.agent_name}_client" in before_agent
            and self.agent.retrieve_config.vector_db == "chroma"
        ):
            args_content += (
                f'        "client": {self.agent_name}_client,' + "\n"
            )
        args_content += closing_arg
        return args_content

    def _get_model_arg(
        self,
    ) -> str:
        if not isinstance(self.agent, WaldiezRagUserProxy):
            return ""
        if self.agent.data.model_ids:
            model_name = self.model_names[self.agent.data.model_ids[0]]
            new_model_name = f"{model_name}"
            return f"{new_model_name}"
        if self.agent.retrieve_config.model in self.model_names:
            selected_model = self.model_names[self.agent.retrieve_config.model]
            new_model_name = f"{selected_model}"
            return f"{new_model_name}"
        return WaldiezRagUserProxyModels[self.agent.retrieve_config.vector_db]

    def _get_args_dict(self) -> dict[str, str | list[str]]:
        if not isinstance(self.agent, WaldiezRagUserProxy):
            return {}
        model_arg = self._get_model_arg()
        args_dict: dict[str, str | list[str]] = {
            "task": self.agent.retrieve_config.task,
            "model": model_arg,
        }
        optional_args = [
            "chunk_token_size",
            "context_max_tokens",
            "customized_prompt",
            "customized_answer_prefix",
        ]
        for arg in optional_args:
            arg_value = getattr(self.agent.retrieve_config, arg)
            if arg_value is not None:
                args_dict[arg] = arg_value
                args_dict[arg] = getattr(self.agent.retrieve_config, arg)
        docs_path: str | list[str] = []
        if self.agent.retrieve_config.docs_path:
            doc_paths = (
                self.agent.retrieve_config.docs_path
                if isinstance(self.agent.retrieve_config.docs_path, list)
                else [self.agent.retrieve_config.docs_path]
            )
            docs_path = [
                item
                for item in [
                    self.path_resolver.resolve(path) for path in doc_paths
                ]
                if item
            ]
            args_dict["docs_path"] = docs_path
        if docs_path:
            args_dict["docs_path"] = docs_path
        non_optional_args = [
            "new_docs",
            "update_context",
            "get_or_create",
            "overwrite",
            "recursive",
            "chunk_mode",
            "must_break_at_empty_line",
            "collection_name",
            "distance_threshold",
        ]
        for arg in non_optional_args:
            args_dict[arg] = getattr(self.agent.retrieve_config, arg)
        return args_dict



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Step-by-step runner for Waldiez workflows."""

from .breakpoints_mixin import BreakpointsMixin
from .step_by_step_models import (
    WaldiezDebugBreakpointAdded,
    WaldiezDebugBreakpointCleared,
    WaldiezDebugBreakpointRemoved,
    WaldiezDebugBreakpointsList,
    WaldiezDebugError,
    WaldiezDebugEventInfo,
    WaldiezDebugHelp,
    WaldiezDebugHelpCommand,
    WaldiezDebugHelpCommandGroup,
    WaldiezDebugInputRequest,
    WaldiezDebugInputResponse,
    WaldiezDebugMessage,
    WaldiezDebugMessageWrapper,
    WaldiezDebugPrint,
    WaldiezDebugStats,
    WaldiezDebugStepAction,
)
from .step_by_step_runner import WaldiezStepByStepRunner

__all__ = [
    "BreakpointsMixin",
    "WaldiezDebugError",
    "WaldiezDebugEventInfo",
    "WaldiezDebugHelp",
    "WaldiezDebugHelpCommand",
    "WaldiezDebugHelpCommandGroup",
    "WaldiezDebugInputRequest",
    "WaldiezDebugInputResponse",
    "WaldiezDebugMessage",
    "WaldiezDebugMessageWrapper",
    "WaldiezDebugPrint",
    "WaldiezDebugStats",
    "WaldiezDebugStepAction",
    "WaldiezStepByStepRunner",
    "WaldiezDebugBreakpointAdded",
    "WaldiezDebugBreakpointCleared",
    "WaldiezDebugBreakpointRemoved",
    "WaldiezDebugBreakpointsList",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Ensure unique names for agents, models, tools, and chats."""

from collections.abc import Iterable
from typing import TypedDict

from ..agents import WaldiezAgent
from ..chat import WaldiezChat
from ..common import MAX_VARIABLE_LENGTH, get_valid_instance_name
from ..model import WaldiezModel
from ..tool import WaldiezTool


class WaldiezUniqueNames(TypedDict):
    """The result type for ensure_unique_names."""

    agent_names: dict[str, str]
    model_names: dict[str, str]
    tool_names: dict[str, str]
    chat_names: dict[str, str]
    agents: list[WaldiezAgent]
    models: list[WaldiezModel]
    tools: list[WaldiezTool]
    chats: list[WaldiezChat]
    flow_name: str


# pylint: disable=too-many-locals
def ensure_unique_names(
    flow_id: str,
    flow_name: str,
    flow_agents: Iterable[WaldiezAgent],
    flow_models: Iterable[WaldiezModel],
    flow_tools: Iterable[WaldiezTool],
    flow_chats: Iterable[WaldiezChat],
    max_length: int = MAX_VARIABLE_LENGTH,
    flow_name_max_length: int = 20,
) -> WaldiezUniqueNames:
    """Ensure unique names for agents, models, tools, and chats and flow.

    Parameters
    ----------
    flow_id : str
        The ID of the flow.
    flow_name : str
        The name of the flow.
    flow_agents : Iterable[WaldiezAgent]
        The agents in the flow.
    flow_models : Iterable[WaldiezModel]
        The models in the flow.
    flow_tools : Iterable[WaldiezTool]
        The tools in the flow.
    flow_chats : Iterable[WaldiezChat]
        The chats in the flow.
    max_length : int, optional
        The maximum length of the name, by default 46
    flow_name_max_length : int, optional
        The maximum length of the flow name, by default 20

    Returns
    -------
    ResultType
        The result with unique names for agents, models, tools, chats, flow.
    """
    all_names: dict[str, str] = {}
    agent_names: dict[str, str] = {}
    model_names: dict[str, str] = {}
    tool_names: dict[str, str] = {}
    chat_names: dict[str, str] = {}
    agents: list[WaldiezAgent] = []
    models: list[WaldiezModel] = []
    tools: list[WaldiezTool] = []
    chats: list[WaldiezChat] = []

    for agent in flow_agents:
        all_names = get_valid_instance_name(
            (agent.id, agent.name),
            all_names,
            prefix="wa",
            max_length=max_length,
        )
        agent_names[agent.id] = all_names[agent.id]
        agents.append(agent)

    for model in flow_models:
        all_names = get_valid_instance_name(
            (model.id, model.name),
            all_names,
            prefix="wm",
            max_length=max_length,
        )
        model_names[model.id] = all_names[model.id]
        models.append(model)

    for tool in flow_tools:
        all_names = get_valid_instance_name(
            (tool.id, tool.name),
            all_names,
            prefix="ws",
            max_length=max_length,
        )
        tool_names[tool.id] = all_names[tool.id]
        tools.append(tool)

    for chat in flow_chats:
        all_names = get_valid_instance_name(
            (chat.id, chat.name), all_names, prefix="wc", max_length=max_length
        )
        chat_names[chat.id] = all_names[chat.id]
        chats.append(chat)

    all_names = get_valid_instance_name(
        (flow_id, flow_name),
        all_names,
        prefix="wf",
        max_length=flow_name_max_length,
    )
    flow_name = all_names[flow_id]

    result: WaldiezUniqueNames = {
        "agent_names": agent_names,
        "model_names": model_names,
        "tool_names": tool_names,
        "chat_names": chat_names,
        "agents": agents,
        "models": models,
        "tools": tools,
        "chats": chats,
        "flow_name": flow_name,
    }
    return result



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=broad-exception-caught,no-member,invalid-name
# pyright: reportUnknownVariableType=false,reportAny=false
# pyright: reportDeprecated=false, reportUnannotatedClassAttribute=false
"""Session management models for WebSocket workflow execution."""

import json
import time
import uuid
from enum import Enum
from pathlib import Path
from typing import Annotated, Any, Literal, Union

from pydantic import BaseModel, ConfigDict, Field, ValidationError


class WorkflowStatus(str, Enum):
    """Workflow execution status."""

    IDLE = "idle"
    STARTING = "starting"
    RUNNING = "running"
    PAUSED = "paused"
    STEP_WAITING = "step_waiting"
    INPUT_WAITING = "input_waiting"
    STOPPING = "stopping"
    COMPLETED = "completed"
    FAILED = "failed"
    CANCELLED = "cancelled"


class ExecutionMode(str, Enum):
    """Workflow execution modes."""

    STANDARD = "standard"
    STEP_BY_STEP = "step_by_step"
    SUBPROCESS = "subprocess"


class BaseWaldiezMessage(BaseModel):
    """Base class for all Waldiez WebSocket messages."""

    message_id: str = Field(
        default_factory=lambda: f"msg_{time.monotonic_ns()}",
    )
    timestamp: float = Field(default_factory=time.time)

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="ignore",
        json_encoders={
            Path: str,
            Enum: lambda v: v.value,
        },
    )


class BaseRequest(BaseWaldiezMessage):
    """Base class for client requests."""


class BaseResponse(BaseWaldiezMessage):
    """Base class for server responses."""

    request_id: str | None = None
    success: bool = True
    error: str | None = None

    @classmethod
    def ok(cls, **kwargs: Any) -> "BaseResponse":
        """Create a successful response.

        Parameters
        ----------
        kwargs : Any
            Additional keyword arguments to include in the response.

        Returns
        -------
        BaseResponse
            The created successful response.

        """
        kwargs.setdefault("success", True)
        kwargs.setdefault("error", None)
        return cls(**kwargs)

    @classmethod
    def fail(cls, error: str, **kwargs: Any) -> "BaseResponse":
        """
        Create a failed response.

        Parameters
        ----------
        error : str
            The error message.
        kwargs : Any
            Additional keyword arguments to include in the response.

        Returns
        -------
        BaseResponse
            The created failed response.

        """
        kwargs.update({"success": False, "error": error})
        return cls(**kwargs)


class BaseNotification(BaseWaldiezMessage):
    """Base class for server notifications (no response expected)."""


# ========================================
# CLIENT-TO-SERVER MESSAGES (REQUESTS)
# ========================================


class SaveFlowRequest(BaseRequest):
    """Request to save a workflow."""

    type: Literal["save"] = "save"
    data: str  # JSON string of workflow
    path: str | None = None
    force: bool = False


class RunWorkflowRequest(BaseRequest):
    """Request to run a workflow in standard mode."""

    type: Literal["run"] = "run"
    data: str  # JSON string of workflow
    path: str | None = None


class StepRunWorkflowRequest(BaseRequest):
    """Request to run a workflow in step-by-step mode."""

    type: Literal["step_run"] = "step_run"
    data: str  # JSON string of workflow
    breakpoints: list[str] = Field(default_factory=list)
    checkpoint: str | None = None
    path: str | None = None


class StepControlRequest(BaseRequest):
    """Request to control step-by-step execution."""

    type: Literal["step_control"] = "step_control"
    action: Literal[
        "continue",
        "step",
        "run",
        "quit",
        "info",
        "help",
        "stats",
        "",
        "c",
        "s",
        "r",
        "q",
        "i",
        "h",
        "?",
        "st",
    ]
    session_id: str


class BreakpointRequest(BaseRequest):
    """Request to manage breakpoints."""

    type: Literal["breakpoint_control"] = "breakpoint_control"
    action: Literal["add", "remove", "list", "clear"]
    event_type: str | None = None  # Required for add/remove
    session_id: str

    @property
    def response(self) -> Literal["added", "removed", "list", "cleared"]:
        """Get the response message for the breakpoint action."""
        if self.action == "remove":
            return "removed"
        if self.action == "list":
            return "list"
        if self.action == "clear":
            return "cleared"
        return "added"


class UserInputResponse(BaseRequest):
    """User input response for workflow execution."""

    type: Literal["user_input"] = "user_input"
    request_id: str
    data: Any
    session_id: str


class StopWorkflowRequest(BaseRequest):
    """Request to stop workflow execution."""

    type: Literal["stop"] = "stop"
    session_id: str
    force: bool = False


class StopWorkflowResponse(BaseResponse):
    """Response to stop workflow request."""

    session_id: str
    type: Literal["stop_response"] = "stop_response"
    error: str | None = None
    forced: bool = False


class ConvertWorkflowRequest(BaseRequest):
    """Request to convert workflow to different format."""

    type: Literal["convert"] = "convert"
    data: str
    format: Literal["py", "ipynb"]
    path: str | None = None


class UploadFileRequest(BaseRequest):
    """Request to upload a file."""

    type: Literal["upload_file"] = "upload_file"
    filename: str
    file_data: str  # Base64 encoded
    file_size: int
    mime_type: str | None = None


class PingRequest(BaseRequest):
    """Ping request for connection testing."""

    type: Literal["ping"] = "ping"
    echo_data: dict[str, Any] = Field(default_factory=dict)


class GetStatusRequest(BaseRequest):
    """Request current server/workflow status."""

    type: Literal["get_status"] = "get_status"
    session_id: str | None = None


class GetCheckpointsRequest(BaseRequest):
    """Request previous checkpoints for a flow."""

    request_id: str
    type: Literal["get_checkpoints"] = "get_checkpoints"
    payload: str | dict[str, Any]


class SetCheckpointRequest(BaseRequest):
    """Save a checkpoint."""

    request_id: str
    type: Literal["set_checkpoint"] = "set_checkpoint"
    payload: str | dict[str, Any]


class DeleteCheckpointRequest(BaseRequest):
    """Delete a checkpoint."""

    request_id: str
    type: Literal["delete_checkpoint"] = "delete_checkpoint"
    payload: str | dict[str, Any]


# ========================================
# SERVER-TO-CLIENT MESSAGES (RESPONSES)
# ========================================


class SaveFlowResponse(BaseResponse):
    """Response to save flow request."""

    type: Literal["save_response"] = "save_response"
    path: str | None = None
    error: str | None = None


class RunWorkflowResponse(BaseResponse):
    """Response to run workflow request."""

    type: Literal["run_response"] = "run_response"
    session_id: str
    error: str | None = None


class StepRunWorkflowResponse(BaseResponse):
    """Response to step run workflow request."""

    type: Literal["step_run_response"] = "step_run_response"
    session_id: str
    breakpoints: list[str]
    checkpoint: str | None = None
    error: str | None = None


class StepControlResponse(BaseResponse):
    """Response to step control request."""

    type: Literal["step_control_response"] = "step_control_response"
    action: str
    result: str
    session_id: str


class BreakpointResponse(BaseResponse):
    """Response to breakpoint management."""

    type: Literal["breakpoint_response"] = "breakpoint_response"
    action: str
    breakpoints: list[str] = Field(default_factory=list)
    session_id: str
    error: str | None = None


class ConvertWorkflowResponse(BaseResponse):
    """Response to convert workflow request."""

    type: Literal["convert_response"] = "convert_response"
    format: str
    path: str | None = None
    error: str | None = None


class UploadFileResponse(BaseResponse):
    """Response to file upload."""

    type: Literal["upload_file_response"] = "upload_file_response"
    path: str | None = None
    size: int | None = None
    error: str | None = None


class PongResponse(BaseResponse):
    """Response to ping."""

    type: Literal["pong"] = "pong"
    echo_data: dict[str, Any] = Field(default_factory=dict)
    server_time: float = Field(default_factory=time.time)


class StatusResponse(BaseResponse):
    """Response with current status."""

    type: Literal["status_response"] = "status_response"
    server_status: dict[str, Any]
    workflow_status: WorkflowStatus | None = None
    session_id: str | None = None


class GetCheckpointsResponse(BaseResponse):
    """Return the checkpoints of a flow."""

    type: Literal["get_checkpoints"] = "get_checkpoints"
    checkpoints: dict[str, Any] = Field(default_factory=dict)


class SetCheckpointResponse(BaseResponse):
    """Save a checkpoint of a flow."""

    type: Literal["set_checkpoint"] = "set_checkpoint"
    checkpoint: dict[str, Any] = Field(default_factory=dict)


class DeleteCheckpointResponse(BaseResponse):
    """Delete a checkpoint of a flow."""

    type: Literal["delete_checkpoint"] = "delete_checkpoint"
    checkpoint: str


class ErrorResponse(BaseResponse):
    """Generic error response."""

    type: Literal["error"] = "error"
    error_code: int
    error_type: str
    details: dict[str, Any] = Field(default_factory=dict)
    success: bool = False


# ========================================
# SERVER-TO-CLIENT NOTIFICATIONS
# ========================================


class WorkflowStatusNotification(BaseNotification):
    """Notification of workflow status change."""

    type: Literal["workflow_status"] = "workflow_status"
    session_id: str
    status: WorkflowStatus
    mode: ExecutionMode
    details: str | None = None

    @classmethod
    def make(
        cls,
        session_id: str,
        status: WorkflowStatus,
        mode: ExecutionMode,
        details: str | None = None,
    ) -> "WorkflowStatusNotification":
        """Create a workflow status notification.

        Parameters
        ----------
        session_id : str
            The session ID.
        status : WorkflowStatus
            The workflow status.
        mode : ExecutionMode
            The execution mode.
        details : str | None
            Additional details about the status.

        Returns
        -------
        WorkflowStatusNotification
            The created workflow status notification.
        """
        return cls(
            session_id=session_id,
            status=status,
            mode=mode,
            details=details,
        )


class WorkflowOutputNotification(BaseNotification):
    """Notification of workflow output."""

    type: Literal["workflow_output"] = "workflow_output"
    session_id: str
    stream: Literal["stdout", "stderr"]
    content: str
    output_type: Literal["text", "structured", "debug"] = "text"

    @classmethod
    def stdout(cls, session_id: str, text: str) -> "WorkflowOutputNotification":
        """Create a workflow output notification for stdout.

        Parameters
        ----------
        session_id : str
            The session ID.
        text : str
            The output text.

        Returns
        -------
        WorkflowOutputNotification
            The created workflow output notification.
        """
        return cls(session_id=session_id, stream="stdout", content=text)

    @classmethod
    def stderr(cls, session_id: str, text: str) -> "WorkflowOutputNotification":
        """Create a workflow output notification for stderr.

        Parameters
        ----------
        session_id : str
            The session ID.
        text : str
            The output text.

        Returns
        -------
        WorkflowOutputNotification
            The created workflow output notification.
        """
        return cls(session_id=session_id, stream="stderr", content=text)


class WorkflowEventNotification(BaseNotification):
    """Notification of workflow event (for step-by-step)."""

    type: Literal["workflow_event"] = "workflow_event"
    session_id: str
    event_data: dict[str, Any]
    event_count: int
    should_break: bool = False


class UserInputRequestNotification(BaseNotification):
    """Notification requesting user input."""

    type: Literal["input_request"] = "input_request"
    session_id: str
    request_id: str
    prompt: str
    password: bool = False
    timeout: float = 120.0


class BreakpointNotification(BaseNotification):
    """Notification about breakpoint changes."""

    type: Literal["breakpoint_notification"] = "breakpoint_notification"
    session_id: str
    action: Literal["added", "removed", "cleared", "list"]
    event_type: str | None = None
    breakpoints: list[str] = Field(default_factory=list)
    message: str | None = None


class WorkflowCompletionNotification(BaseNotification):
    """Notification of workflow completion."""

    type: Literal["workflow_completion"] = "workflow_completion"
    session_id: str
    success: bool
    exit_code: int | None = None
    results: list[dict[str, Any]] = Field(default_factory=list)
    execution_time: float | None = None
    error: str | None = None


class StepDebugNotification(BaseNotification):
    """Notification for step-by-step debug information."""

    type: Literal["step_debug"] = "step_debug"
    session_id: str
    debug_type: Literal["stats", "help", "error", "info"]
    data: dict[str, Any]


class ConnectionNotification(BaseNotification):
    """Notification about connection status."""

    type: Literal["connection"] = "connection"
    status: Literal["connected", "disconnected", "error"]
    client_id: str
    server_time: float = Field(default_factory=time.time)
    message: str | None = None


# ========================================
# SUBPROCESS-SPECIFIC MODELS
# ========================================


class SubprocessOutputNotification(BaseNotification):
    """Notification from subprocess execution."""

    type: Literal["subprocess_output"] = "subprocess_output"
    session_id: str
    stream: Literal["stdout", "stderr"]
    content: str
    subprocess_type: Literal["output", "error", "debug"] = "output"
    context: dict[str, Any] = Field(default_factory=dict)


class SubprocessInputRequestNotification(BaseNotification):
    """Input request from subprocess."""

    type: Literal["subprocess_input_request"] = "subprocess_input_request"
    session_id: str
    request_id: str
    prompt: str
    timeout: float = 120.0
    password: bool = False
    context: dict[str, Any] = Field(default_factory=dict)


class SubprocessCompletionNotification(BaseNotification):
    """Subprocess execution completion."""

    type: Literal["subprocess_completion"] = "subprocess_completion"
    session_id: str
    success: bool
    exit_code: int
    message: str
    context: dict[str, Any] = Field(default_factory=dict)


# ========================================
# UNION TYPES FOR MESSAGE PARSING
# ========================================

# Client-to-Server Messages
ClientMessage = Annotated[
    Union[
        SaveFlowRequest,
        RunWorkflowRequest,
        StepRunWorkflowRequest,
        StepControlRequest,
        BreakpointRequest,
        UserInputResponse,
        StopWorkflowRequest,
        ConvertWorkflowRequest,
        UploadFileRequest,
        PingRequest,
        GetStatusRequest,
        GetCheckpointsRequest,
        SetCheckpointRequest,
        DeleteCheckpointRequest,
    ],
    Field(discriminator="type"),
]

# Server-to-Client Messages
ServerMessage = Annotated[
    Union[
        # Responses
        SaveFlowResponse,
        RunWorkflowResponse,
        StepRunWorkflowResponse,
        StopWorkflowResponse,
        StepControlResponse,
        BreakpointResponse,
        BreakpointNotification,
        ConvertWorkflowResponse,
        UploadFileResponse,
        PongResponse,
        StatusResponse,
        GetCheckpointsResponse,
        SetCheckpointResponse,
        DeleteCheckpointResponse,
        ErrorResponse,
        # Notifications
        WorkflowStatusNotification,
        WorkflowOutputNotification,
        WorkflowEventNotification,
        UserInputRequestNotification,
        WorkflowCompletionNotification,
        StepDebugNotification,
        ConnectionNotification,
        SubprocessOutputNotification,
        SubprocessInputRequestNotification,
        SubprocessCompletionNotification,
    ],
    Field(discriminator="type"),
]


class _ServerMessageWrapper(BaseModel):
    """Wrapper for server messages to handle discriminators."""

    # noinspection PyTypeHints
    message: ServerMessage


class _ClientMessageWrapper(BaseModel):
    """Wrapper for client messages to handle discriminators."""

    # noinspection PyTypeHints
    message: ClientMessage


# All messages
WaldiezWsMessage = Annotated[
    Union[ClientMessage, ServerMessage],
    Field(discriminator="type"),
]


# ========================================
# UTILITY FUNCTIONS
# ========================================


# noinspection PyTypeHints,DuplicatedCode
def parse_client_message(data: str | dict[str, Any]) -> ClientMessage:
    """Parse client message from JSON string or dict.

    Parameters
    ----------
    data : str | dict[str, Any]
        Message data

    Returns
    -------
    ClientMessage
        Parsed client message

    Raises
    ------
    ValueError
        If message cannot be parsed
    """
    if isinstance(data, str):
        try:
            data = json.loads(data)
        except json.JSONDecodeError as e:
            raise ValueError(f"Invalid message format: {e}") from e
    try:
        return _ClientMessageWrapper.model_validate({"message": data}).message
    except ValidationError as e:
        raise ValueError(f"Invalid client message: {e}") from e


# noinspection PyTypeHints,DuplicatedCode
def parse_server_message(data: str | dict[str, Any]) -> ServerMessage:
    """Parse server message from JSON string or dict.

    Parameters
    ----------
    data : str | dict[str, Any]
        Message data

    Returns
    -------
    ServerMessage
        Parsed server message

    Raises
    ------
    ValueError
        If message cannot be parsed
    """
    if isinstance(data, str):
        try:
            data = json.loads(data)
        except json.JSONDecodeError as e:
            raise ValueError(f"Invalid message format: {e}") from e
    try:
        return _ServerMessageWrapper.model_validate({"message": data}).message
    except ValidationError as e:
        raise ValueError(f"Invalid server message: {e}") from e


def create_error_response(
    error_message: str,
    error_code: int = 500,
    error_type: str = "InternalError",
    request_id: str | None = None,
    details: dict[str, Any] | None = None,
) -> ErrorResponse:
    """Create standardized error response.

    Parameters
    ----------
    error_message : str
        Error message
    error_code : int
        Error code
    error_type : str
        Error type
    request_id : str | None
        Request ID if this is response to a request
    details : dict[str, Any] | None
        Additional error details

    Returns
    -------
    ErrorResponse
        Standardized error response
    """
    return ErrorResponse(
        request_id=request_id,
        error_code=error_code,
        error_type=error_type,
        error=error_message,
        details=details or {},
    )


def create_session_id() -> str:
    """Create a new session ID.

    Returns
    -------
    str
        New session ID
    """
    return f"session_{uuid.uuid4().hex[:12]}"


# ========================================
# MESSAGE TYPE CONSTANTS
# ========================================

CLIENT_MESSAGE_TYPES = {
    "save",
    "run",
    "step_run",
    "step_control",
    "breakpoint_control",
    "user_input",
    "stop",
    "convert",
    "upload_file",
    "ping",
    "get_status",
}

SERVER_MESSAGE_TYPES = {
    # Responses
    "save_response",
    "run_response",
    "step_run_response",
    "step_control_response",
    "breakpoint_response",
    "convert_response",
    "upload_file_response",
    "pong",
    "status_response",
    "error",
    # Notifications
    "workflow_status",
    "workflow_output",
    "workflow_event",
    "input_request",
    "breakpoint_notification",
    "workflow_completion",
    "step_debug",
    "connection",
    "subprocess_output",
    "subprocess_input_request",
    "subprocess_completion",
}

ALL_MESSAGE_TYPES = CLIENT_MESSAGE_TYPES | SERVER_MESSAGE_TYPES


class SessionState(BaseModel):
    """State information for a workflow session."""

    session_id: str
    client_id: str
    status: WorkflowStatus
    mode: ExecutionMode
    start_time: int = Field(default_factory=time.monotonic_ns)
    end_time: int | None = None
    metadata: dict[str, Any] = Field(default_factory=dict)

    # Runtime fields (not serialized)
    runner: Any = Field(default=None, exclude=True)
    temp_file: Path | None = Field(default=None, exclude=True)

    @property
    def duration(self) -> float:
        """Get session duration in seconds."""
        end = self.end_time or time.monotonic_ns()
        return (end - self.start_time) / 1_000_000_000

    @property
    def is_active(self) -> bool:
        """Check if session is currently active."""
        return self.status in {
            WorkflowStatus.STARTING,
            WorkflowStatus.RUNNING,
            WorkflowStatus.PAUSED,
            WorkflowStatus.STEP_WAITING,
            WorkflowStatus.INPUT_WAITING,
        }

    @property
    def is_completed(self) -> bool:
        """Check if session has completed (successfully or not)."""
        return self.status in {
            WorkflowStatus.COMPLETED,
            WorkflowStatus.FAILED,
            WorkflowStatus.CANCELLED,
        }

    def update_status(self, new_status: WorkflowStatus) -> None:
        """Update session status and set end time if completed.

        Parameters
        ----------
        new_status : WorkflowStatus
            The new status to set.
        """
        self.status = new_status
        if self.is_completed and not self.end_time:
            self.end_time = time.monotonic_ns()

    def get_execution_summary(self) -> dict[str, Any]:
        """Get a summary of session execution.

        Returns
        -------
        dict[str, Any]
            A dictionary containing session execution summary.
        """
        return {
            "session_id": self.session_id,
            "client_id": self.client_id,
            "status": self.status.value,
            "mode": self.mode.value,
            "duration_seconds": self.duration,
            "start_time": self.start_time,
            "end_time": self.end_time,
            "is_active": self.is_active,
            "is_completed": self.is_completed,
        }



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pyright: reportMissingTypeStubs=false
"""Common utilities for the waldiez runner."""

import asyncio
import sys
import traceback
from dataclasses import dataclass
from getpass import getpass
from pathlib import Path
from typing import Any, Callable


@dataclass
class ProcessSetup:
    """Container for subprocess setup data."""

    temp_dir: Path
    file_path: Path
    old_vars: dict[str, str]
    skip_mmd: bool


async def input_async(prompt: str, *, password: bool = False) -> str:
    """Asynchronous input function.

    Parameters
    ----------
    prompt : str
        The prompt to display to the user.
    password : bool, optional
        Whether to hide input (password mode), by default False.

    Returns
    -------
    str
        The user input.
    """
    if password:
        try:
            return await asyncio.to_thread(getpass, prompt)
        except EOFError:
            return ""
    try:
        return await asyncio.to_thread(input, prompt)
    except EOFError:
        return ""


def input_sync(prompt: str, *, password: bool = False) -> str:
    """Input function (synchronous).

    Parameters
    ----------
    prompt : str
        The prompt to display to the user.
    password : bool, optional
        Whether to hide input (password mode), by default False.

    Returns
    -------
    str
        The user input.
    """
    if password:
        try:
            return getpass(prompt)
        except EOFError:
            return ""
    try:
        return input(prompt)
    except EOFError:
        return ""


# pylint: disable=import-outside-toplevel,too-complex
def get_printer() -> Callable[..., None]:  # noqa: C901
    """Get the printer function.

    Returns
    -------
    Callable[..., None]
        The printer function that handles Unicode encoding errors gracefully.
    """
    try:
        # noinspection PyUnresolvedReferences
        from autogen.io import IOStream  # type: ignore

        printer = IOStream.get_default().print
    except ImportError:  # pragma: no cover
        # Fallback to standard print if autogen is not available
        printer = print

    # noinspection PyBroadException,TryExceptPass
    def safe_printer(*args: Any, **kwargs: Any) -> None:  # noqa: C901
        """Safe printer that handles Unicode encoding errors.

        Parameters
        ----------
        *args : Any
            Arguments to pass to the printer
        **kwargs : Any
            Keyword arguments to pass to the printer
        """
        # pylint: disable=broad-exception-caught,too-many-try-statements
        try:
            printer(*args, **kwargs)
        except (UnicodeEncodeError, UnicodeDecodeError):
            # First fallback: try to get a safe string representation
            try:
                msg, flush = get_what_to_print(*args, **kwargs)
                # Convert problematic characters to safe representations
                safe_msg = msg.encode("utf-8", errors="replace").decode("utf-8")
                printer(safe_msg, end="", flush=flush)
            except (UnicodeEncodeError, UnicodeDecodeError):
                # Second fallback: use built-in print with safe encoding
                try:
                    # Convert args to safe string representations
                    safe_args: list[str] = []
                    for arg in args:
                        try:
                            safe_args.append(
                                str(arg)
                                .encode("utf-8", errors="replace")
                                .decode("utf-8")
                            )
                        except (
                            UnicodeEncodeError,
                            UnicodeDecodeError,
                        ):  # pragma: no cover
                            safe_args.append(repr(arg))

                    # Use built-in print instead of the custom printer
                    print(*safe_args, **kwargs)

                except Exception:
                    # Final fallback: write directly to stderr buffer
                    error_msg = (
                        "Could not print the message due to encoding issues.\n"
                    )
                    to_sys_stderr(error_msg)
        except Exception as e:
            # Handle any other unexpected errors
            traceback.print_exc()
            error_msg = f"Unexpected error in printer: {str(e)}\n"
            to_sys_stderr(error_msg)

    return safe_printer


def to_sys_stderr(msg: str) -> None:
    """Write a message to sys.stderr.

    Parameters
    ----------
    msg : str
        The message to write to stderr.
    """
    # pylint: disable=broad-exception-caught
    # noinspection TryExceptPass,PyBroadException
    try:
        if hasattr(sys.stderr, "buffer"):
            sys.stderr.buffer.write(msg.encode("utf-8", errors="replace"))
            sys.stderr.buffer.flush()
        else:  # pragma: no cover
            sys.stderr.write(msg)
            sys.stderr.flush()
    except Exception:  # pragma: no cover
        pass


def get_what_to_print(*args: Any, **kwargs: Any) -> tuple[str, bool]:
    """Extract message and flush flag from print arguments.

    Parameters
    ----------
    *args : Any
        Arguments to print
    **kwargs : Any
        Keyword arguments for print function

    Returns
    -------
    tuple[str, bool]
        Message to print and flush flag
    """
    # Convert all args to strings and join with spaces (like print does)
    msg = " ".join(str(arg) for arg in args)

    # Handle sep parameter
    sep = kwargs.get("sep", " ")
    if isinstance(sep, bytes):  # pragma: no cover
        sep = sep.decode("utf-8", errors="replace")
    if len(args) > 1:
        msg = sep.join(str(arg) for arg in args)

    # Handle end parameter
    end = kwargs.get("end", "\n")
    if isinstance(end, bytes):  # pragma: no cover
        end = end.decode("utf-8", errors="replace")
    msg += end

    # Handle flush parameter
    flush = kwargs.get("flush", False)
    # noinspection PyUnreachableCode
    if not isinstance(flush, bool):  # pragma: no cover
        flush = False

    return msg, flush



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=too-few-public-methods,no-self-use
"""Code execution processor for Waldiez agents."""

import json

from typing_extensions import Literal

from waldiez.models import WaldiezAgent, WaldiezAgentCodeExecutionConfig

from ..core import CodeExecutionConfig, ImportPosition, ImportStatement


class CodeExecutionProcessor:
    """Processor for code execution configuration."""

    def __init__(
        self,
        agent: WaldiezAgent,
        agent_name: str,
        tool_names: dict[str, str],
    ):
        self.agent = agent
        self.agent_name = agent_name
        self.tool_names = tool_names

    def process(self) -> CodeExecutionConfig:
        """Process code execution configuration.

        Returns
        -------
        CodeExecutionConfig
            The processed code execution configuration.
        """
        if self.agent.data.code_execution_config is False:
            return CodeExecutionConfig(
                executor_content="",
                executor_argument="False",
                executor_import=None,
            )

        config = self.agent.data.code_execution_config
        use_docker = (
            config.use_docker if config.use_docker is not None else False
        )

        executor_class = self._get_executor_class_name(use_docker)
        executor_content = self._build_executor_content(config, use_docker)
        executor_arg = f'{{"executor": {self.agent_name}_executor}}'
        executor_import = f"from autogen.coding import {executor_class}"

        return CodeExecutionConfig(
            executor_content=executor_content,
            executor_argument=executor_arg,
            executor_import=ImportStatement(
                statement=executor_import,
                position=ImportPosition.THIRD_PARTY,
            ),
        )

    # noinspection PyMethodMayBeStatic
    def _get_executor_class_name(self, use_docker: bool) -> str:
        """Get the appropriate executor class name."""
        return (
            "DockerCommandLineCodeExecutor"
            if use_docker
            else "LocalCommandLineCodeExecutor"
        )

    def _build_executor_content(
        self,
        config: WaldiezAgentCodeExecutionConfig | Literal[False],
        use_docker: bool,
    ) -> str:
        """Build the executor content string."""
        if config is False:
            return ""
        executor_class = self._get_executor_class_name(use_docker)
        lines = [f"{self.agent_name}_executor = {executor_class}("]

        # Add work directory
        if config.work_dir:
            lines.append(f"    work_dir={json.dumps(config.work_dir)},")

        # Add timeout
        if config.timeout:
            lines.append(f"    timeout={int(config.timeout)},")

        # Add functions (only for local executor)
        if not use_docker and config.functions:
            function_names = self._get_function_names(config.functions)
            if function_names:
                function_names_str = ", ".join(function_names)
                lines.append(f"    functions=[{function_names_str}],")

        lines.append(")")
        lines.append("")  # Add empty line

        return "\n".join(lines)

    def _get_function_names(self, function_ids: list[str]) -> list[str]:
        """Get function names from function IDs.

        Parameters
        ----------
        function_ids : list[str]
            List of function IDs to resolve names for.

        Returns
        -------
        list[str]
            List of function names corresponding to the provided IDs.
        """
        function_names: list[str] = []
        for tool_id in function_ids:
            if tool_id in self.tool_names:
                tool_name = self.tool_names[tool_id]
                if tool_name:
                    function_names.append(tool_name)
        return function_names



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pylint: disable=unused-argument,no-self-use
# pyright: reportMissingTypeStubs=false, reportImportCycles=false
# pyright: reportDeprecated=false, reportUnknownMemberType=false
# pyright: reportUnknownVariableType=false, reportUnknownArgumentType=false

"""Command handler for step-by-step execution."""

import inspect
from collections.abc import Mapping, Sequence
from dataclasses import asdict, is_dataclass
from typing import TYPE_CHECKING, Any, Optional, Union

if TYPE_CHECKING:
    from autogen.agentchat import ConversableAgent  # type: ignore
    from autogen.events.base_event import BaseEvent  # type: ignore
    from autogen.messages.base_message import BaseMessage  # type: ignore

    # noinspection PyUnusedImports
    from .step_by_step_runner import WaldiezStepByStepRunner


# pylint: disable=too-few-public-methods
# noinspection PyBroadException
class EventProcessor:
    """Processes events for the step-by-step runner."""

    def __init__(self, runner: "WaldiezStepByStepRunner"):
        self.runner = runner

    def process_event(
        self,
        event: Union["BaseEvent", "BaseMessage"],
        agents: list["ConversableAgent"],
    ) -> dict[str, Any]:
        """Shared logic for both sync and async event processing.

        Parameters
        ----------
        event : BaseEvent | BaseMessage
            The event to process.
        agents : list[ConversableAgent]
            The workflow's known agents.

        Returns
        -------
        dict[str, Any]
            The result of processing the event.
        """
        self.runner.event_plus_one()
        self.runner.current_event = event

        if self.runner.is_stop_requested():
            return {"action": "stop", "reason": "stop_requested"}

        event_info = self._create_event_info(event)
        self._update_participant_info(event_info)
        self._manage_event_history(event_info)
        self._check_for_input_request(event_info)
        self._add_agents_info(event_info, agents)

        should_break = self.runner.should_break_on_event(
            event, sender_only=True
        )

        return {
            "action": "break" if should_break else "continue",
            "event_info": event_info,
        }

    def _create_event_info(
        self,
        event: Union["BaseEvent", "BaseMessage"],
    ) -> dict[str, Any]:
        """Create event info dictionary from event object.

        Parameters
        ----------
        event : BaseEvent | BaseMessage
            The event to convert to info dict.

        Returns
        -------
        dict[str, Any]
            Event information dictionary.
        """
        event_info = event.model_dump(
            mode="json", exclude_none=True, fallback=str
        )
        event_info["count"] = self.runner.event_count
        event_info["sender"] = getattr(event, "sender", self.runner.last_sender)
        event_info["recipient"] = getattr(
            event, "recipient", self.runner.last_recipient
        )
        return event_info

    def _update_participant_info(self, event_info: dict[str, Any]) -> None:
        """Update sender and recipient information in event_info.

        Parameters
        ----------
        event_info : dict[str, Any]
            Event information dictionary to update.
        """
        if not event_info["sender"] or not event_info["recipient"]:
            self._extract_participants_from_content(event_info)

        self._check_for_event_speaker(event_info)
        self._extract_participants_from_direct_content(event_info)

        # Update last known participants
        self.runner.last_sender = event_info["sender"]
        self.runner.last_recipient = event_info["recipient"]

    @staticmethod
    def _extract_participants_from_content(event_info: dict[str, Any]) -> None:
        """Extract sender/recipient from nested content structure.

        Parameters
        ----------
        event_info : dict[str, Any]
            Event information dictionary to update.
        """
        content = event_info.get("content", {})
        if (
            isinstance(content, dict)
            and "chat_info" in content
            and isinstance(content["chat_info"], dict)
        ):
            content = content.get("chat_info", {})

        if not event_info["sender"] and "sender" in content:
            event_info["sender"] = content["sender"]
        if not event_info["recipient"] and "recipient" in content:
            event_info["recipient"] = content["recipient"]

    @staticmethod
    def _check_for_event_speaker(event_info: dict[str, Any]) -> None:
        """Handle speaker information for group chat events.

        Parameters
        ----------
        event_info : dict[str, Any]
            Event information dictionary to update.
        """
        if "content" in event_info and isinstance(event_info["content"], dict):
            content = event_info.get("content", {})
            speaker = content.get("speaker")
            if isinstance(speaker, str) and speaker:
                event_info["sender"] = speaker

    @staticmethod
    def _extract_participants_from_direct_content(
        event_info: dict[str, Any],
    ) -> None:
        """Extract sender/recipient directly from content dictionary.

        Parameters
        ----------
        event_info : dict[str, Any]
            Event information dictionary to update.
        """
        if "content" in event_info and isinstance(event_info["content"], dict):
            content = event_info.get("content", {})

            sender = content.get("sender", "")
            if isinstance(sender, str) and sender:
                event_info["sender"] = sender

            recipient = content.get("recipient", "")
            if isinstance(recipient, str) and recipient:
                event_info["recipient"] = recipient

    def _manage_event_history(self, event_info: dict[str, Any]) -> None:
        """Add event to history and manage history size limits.

        Parameters
        ----------
        event_info : dict[str, Any]
            Event information to add to history.
        """
        self.runner.add_to_history(event_info)
        self._trim_history_if_needed()

    def _trim_history_if_needed(self) -> None:
        """Remove oldest events from history if over size limit."""
        current_history = self.runner.event_history
        if len(current_history) > self.runner.max_event_history:
            excess = len(current_history) - self.runner.max_event_history
            for _ in range(excess):
                self.runner.pop_event()

    def _check_for_input_request(self, event_info: dict[str, Any]) -> None:
        """Swap participant names if we have an input request."""
        if (
            event_info["type"] in ("input_request", "debug_input_request")
            and "sender" in event_info
            and "recipient" in event_info
        ):
            # swap them,
            # before:
            # "recipient" is the user (the one received the input request),
            # make her the "sender" (the one typing...)
            sender = event_info["sender"]
            recipient = event_info["recipient"]
            event_info["sender"] = recipient
            event_info["recipient"] = sender
            self.runner.last_sender = recipient
            self.runner.last_recipient = sender

    @staticmethod
    def _get_agent_dump(
        agent: Optional["ConversableAgent"],
    ) -> dict[str, Any] | None:
        if not agent:
            return None
        dump = {
            name: _trimmed(value)
            for name, value in vars(agent).items()
            if not name.startswith("_")
            and not inspect.ismethod(value)
            and not inspect.isfunction(value)
            and name.upper() != name
            and name != "chat_messages"
        }
        for attr_key in ["name", "description", "system_message"]:
            attr_val = getattr(agent, attr_key)
            if attr_val:
                dump[attr_key] = _trimmed(attr_val)
        try:
            dump["chat_messages"] = {
                _to_json_key(agent): [_trimmed(message) for message in messages]
                for _agent, messages in agent.chat_messages.items()
            }
        except BaseException:  # pylint: disable=broad-exception-caught
            pass
        dump["cost"] = {
            "actual": agent.get_actual_usage(),
            "total": agent.get_total_usage(),
        }
        return dump

    def _add_agents_info(
        self, event_info: dict[str, Any], agents: list["ConversableAgent"]
    ) -> None:
        """Add agents info."""
        sender = event_info.get("sender", self.runner.last_sender)
        recipient = event_info.get("recipient", self.runner.last_recipient)

        agent_map = {agent.name: agent for agent in agents}

        sender_agent = agent_map.get(sender)
        recipient_agent = agent_map.get(recipient)

        ordered_agents: list["ConversableAgent"] = []
        seen: set[str] = set()

        for a in (sender_agent, recipient_agent):
            if a and a.name not in seen:
                ordered_agents.append(a)
                seen.add(a.name)

        for a in agents:
            if a.name not in seen:
                ordered_agents.append(a)
                seen.add(a.name)

        event_info["agents"] = {
            "sender": self._get_agent_dump(sender_agent),
            "recipient": self._get_agent_dump(recipient_agent),
            "all": [self._get_agent_dump(a) for a in ordered_agents if a],
        }


# pylint: disable=too-complex,too-many-return-statements,
# pylint: disable=broad-exception-caught,too-complex,too-many-branches
# noinspection TryExceptPass,PyBroadException
def _trimmed(  # noqa: C901
    value: Any, max_len: int = 200, _depth: int = 0, _max_depth: int = 10
) -> Any:
    """Recursively trim values for serialization."""
    if _depth >= _max_depth:
        # Hard stop
        s = str(value)
        return (s[:max_len] + "...") if len(s) > max_len else s

    if isinstance(value, str):
        return value[:max_len] + "..." if len(value) > max_len else value

    if value is None or isinstance(value, (int, float, bool)):
        return value

    if isinstance(value, (bytes, bytearray)):
        try:
            s = value.decode("utf-8", errors="replace")
            return s[:max_len] + "..." if len(s) > max_len else s
        except Exception:
            return f"<{type(value).__name__}> {len(value)} bytes"

    if callable(value):
        # pylint: disable=too-many-try-statements
        try:
            name = getattr(value, "__name__", repr(value))
            module = getattr(value, "__module__", "")
            if module:
                s = f"<callable> {module}.{name}"
            else:
                s = f"<callable> {name}"
            return s[:max_len] + "..." if len(s) > max_len else s
        except Exception:
            return "<callable>"

    if _is_dataclass_instance(value):
        try:
            value = asdict(value)
        except Exception:
            s = str(value)
            return (s[:max_len] + "...") if len(s) > max_len else s

    if isinstance(value, Mapping):
        out: dict[str, Any] = {}
        for k, v in value.items():
            jk = _to_json_key(k, max_len)
            out[jk] = _trimmed(v, max_len, _depth + 1, _max_depth)
        return out
    if isinstance(value, (set, frozenset)):
        items = [_trimmed(it, max_len, _depth + 1, _max_depth) for it in value]
        try:
            items.sort(key=repr)
        except Exception:
            pass
        return items

    if isinstance(value, Sequence) and not isinstance(
        value, (str, bytes, bytearray)
    ):
        return [_trimmed(it, max_len, _depth + 1, _max_depth) for it in value]
    if hasattr(value, "__dict__"):
        try:
            return _trimmed(
                {
                    k: v
                    for k, v in vars(value).items()
                    if not str(k).startswith("_") and str(k) != "key"
                },
                max_len,
                _depth + 1,
                _max_depth,
            )
        except Exception:
            pass

    # Fallback: stringify and trim
    try:
        s = str(value)
        return s[:max_len] + "..." if len(s) > max_len else s
    except Exception:
        return None


def _is_dataclass_instance(obj: Any) -> bool:
    """Check if an object is an instance of a dataclass."""
    return is_dataclass(obj) and not isinstance(obj, type)


def _to_json_key(key: Any, max_len: int = 200) -> str:
    """Convert dict key to a JSON-legal key."""
    if key is None or isinstance(key, (bool, int, float)):
        return str(key)

    if isinstance(key, str):
        return key[:max_len] + "..." if len(key) > max_len else key

    for _key in ["name", "_name", "__name__"]:
        if hasattr(key, _key):
            return str(getattr(key, _key))

    # Non-primitive key: include type to avoid collisions like 1 vs "1"
    s = f"<{type(key).__name__}> {repr(key)}"
    return s[:max_len] + "..." if len(s) > max_len else s



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pylint: disable=broad-exception-caught,too-many-try-statements,too-many-locals
# pylint: disable=too-many-positional-arguments,too-many-arguments
# pyright: reportUnknownVariableType=false, reportUnknownMemberType=false
# pyright: reportUnknownArgumentType=false, reportUnusedParameter=false

"""Waldiez run results module."""

import json
import shutil
from collections.abc import Iterable
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, TypedDict

import aiofiles
import anyio.to_thread

from waldiez.storage import StorageManager, safe_name

from .db_utils import a_get_sqlite_out, get_sqlite_out
from .gen_seq_diagram import generate_sequence_diagram
from .io_utils import get_printer
from .post_run import (
    a_get_results_from_json,
    a_store_full_results,
    ensure_error_json as _ensure_error_json,
    get_results_from_json,
    remove_results_json,
    store_full_results,
)
from .timeline_processor import TimelineProcessor


class WaldiezRunResults(TypedDict):
    """Results of the Waldiez run."""

    results: list[dict[str, Any]]
    exception: BaseException | None
    completed: bool


# noinspection PyBroadException,PyUnusedLocal
class ResultsMixin:
    """Results related static methods."""

    RUN_DETAILS = "run.json"

    @staticmethod
    def _cleanup(
        file_path: Path | None = None, tmp_dir: Path | None = None
    ) -> None:
        if tmp_dir and tmp_dir.is_dir():
            try:
                shutil.rmtree(tmp_dir)
            except BaseException:
                pass
        if file_path is None:
            file_path = StorageManager.default_root() / ResultsMixin.RUN_DETAILS
        if not file_path.exists():
            return
        try:
            file_path.unlink()
        except BaseException:
            pass

    @staticmethod
    def _load_from_tmp(file_path: Path) -> tuple[Path, Path, str] | None:
        try:
            with open(file_path, "r", encoding="utf-8") as f:
                details_dict = json.load(f)
        except BaseException:
            ResultsMixin._cleanup(file_path)
            return None
        try:
            src = details_dict.get("src", "")
            dst = details_dict.get("dst", "")
            if not src or not dst:
                ResultsMixin._cleanup(file_path)
                return None
            src_path = Path(src)
            dst_path = Path(dst)
        except BaseException:
            ResultsMixin._cleanup(file_path)
            return None
        if not src_path.is_dir():
            ResultsMixin._cleanup(file_path)
            return None
        try:
            dst_path.parent.mkdir(parents=True, exist_ok=True)
        except BaseException:
            ResultsMixin._cleanup(file_path)
            return None
        flow_name = details_dict.get("name", "")
        if not flow_name or not isinstance(flow_name, str):
            flow_name = dst_path.with_suffix("").name
        return src_path, dst_path, flow_name

    @staticmethod
    def gather() -> tuple[bool, str]:
        """Gather what we have from tmp dir.

        To be used when the flow is interrupted and the tmp dir is not removed.

        Returns
        -------
        tuple[bool, str]
            True if operation succeeded and any related message.
        """
        dir_path = StorageManager.default_root()
        details_path = dir_path / ResultsMixin.RUN_DETAILS

        if not details_path.exists():
            return False, f"No run details found at {details_path}."
        loaded = ResultsMixin._load_from_tmp(details_path)
        if not loaded:
            return False, f"Could not load details from {details_path}."
        tmp_dir, output_file, flow_name = loaded
        waldiez_file = output_file.with_suffix(".waldiez")
        ResultsMixin.post_run(
            results=[],
            error=None,
            temp_dir=tmp_dir,
            output_file=output_file,
            flow_name=flow_name,
            waldiez_file=waldiez_file,
            skip_mmd=True,
            skip_timeline=True,
        )
        return True, f"Gathered run data from {tmp_dir} to {output_file.parent}"

    # pylint: disable=unused-argument
    @staticmethod
    def post_run(
        results: list[dict[str, Any]],
        error: BaseException | None,
        temp_dir: Path,
        output_file: str | Path | None,
        flow_name: str,
        waldiez_file: Path,
        uploads_root: Path | None = None,
        skip_mmd: bool = False,
        skip_timeline: bool = False,
        storage_manager: StorageManager | None = None,
        metadata: dict[str, Any] | None = None,
        copy_artifacts_into: str | None = None,
        keep_tmp: bool = False,
        link_latest: bool = True,
        promote_to_output: Iterable[str] = (
            "tree_of_thoughts.png",
            "reasoning_tree.json",
        ),
        ignore_names: Iterable[str] = (".cache", ".env"),
    ) -> Path | None:
        """Actions to perform after running the flow.

        Parameters
        ----------
        results : list[dict[str, Any]]
            The results of the flow run.
        error : BaseException | None
            Optional error during the run.
        temp_dir : Path
            The temporary directory.
        output_file : str | Path | None, optional
            The output file.
        flow_name : str
            The flow name.
        waldiez_file : Path
            The path of the waldiez file used (or dumped) for the run.
        uploads_root : Path | None, optional
            The runtime uploads root, by default None
        skip_mmd : bool, optional
            Whether to skip the mermaid sequence diagram generation,
            by default, False
        skip_timeline : bool, optional
            Whether to skip the timeline processing, by default False
        storage_manager: StorageManager | None
            The storage manager for copying results and generating symlinks.
        metadata : dict[str, Any] | None
            Optional metadata to store in the checkpoint.
        link_latest : bool
            Whether to update a `latest` link under link_root/session_name.
            Defaults to True.
        keep_tmp : bool
            If False, delete the tmp_dir after copying. Defaults to False.
        copy_artifacts_into : str | None
            If set, copy artifacts into
            checkpoint/<copy_into_subdir> instead of the root.
        promote_to_output : Iterable[str]
            File names (exact matches) to also copy into output_dir.
        ignore_names : Iterable[str]
            Directory/file names to skip entirely.

        Returns
        -------
        Path | None
            The destination directory if output file, else None
        """
        if isinstance(output_file, str):
            output_file = Path(output_file)
        mmd_dir = output_file.parent if output_file else Path.cwd()
        ResultsMixin.ensure_db_outputs(temp_dir)
        if error is not None:
            ResultsMixin.ensure_error_json(temp_dir, error)
        else:
            ResultsMixin.ensure_results_json(temp_dir, results)
        if not skip_mmd:
            ResultsMixin.make_mermaid_diagram(
                temp_dir=temp_dir,
                output_file=output_file,
                flow_name=flow_name,
                mmd_dir=mmd_dir,
            )
        if not skip_timeline:  # pragma: no branch
            ResultsMixin.make_timeline_json(temp_dir)
        if storage_manager is None:
            storage_manager = StorageManager()
        link_root = (
            (output_file.parent / "waldiez_out")
            if output_file
            else (waldiez_file.parent / "waldiez_out")
        )
        output_hint = (
            output_file if output_file else Path.cwd() / waldiez_file.name
        )
        session_name = safe_name(flow_name)
        to_ignore = list(ignore_names)
        if ResultsMixin.RUN_DETAILS not in to_ignore:
            to_ignore.append(ResultsMixin.RUN_DETAILS)
        _checkpoint_path, public_link_path = storage_manager.finalize(
            session_name=session_name,
            output_file=output_hint,
            tmp_dir=temp_dir,
            metadata=metadata or {},
            timestamp=datetime.now(timezone.utc),
            link_root=link_root,
            link_latest=link_latest,
            keep_tmp=keep_tmp,
            copy_into_subdir=copy_artifacts_into,
            promote_to_output=promote_to_output,
            ignore_names=to_ignore,
        )
        try:
            dst_waldiez = public_link_path / waldiez_file.name
            if not dst_waldiez.exists() and waldiez_file.is_file():
                shutil.copyfile(waldiez_file, dst_waldiez)
        except BaseException:
            pass
        ResultsMixin._cleanup(None, None if keep_tmp else temp_dir)
        return public_link_path if output_file else None

    @staticmethod
    async def a_post_run(
        results: list[dict[str, Any]],
        error: BaseException | None,
        temp_dir: Path,
        output_file: str | Path | None,
        flow_name: str,
        waldiez_file: Path,
        uploads_root: Path | None = None,
        skip_mmd: bool = False,
        skip_timeline: bool = False,
        storage_manager: StorageManager | None = None,
        metadata: dict[str, Any] | None = None,
        copy_artifacts_into: str | None = None,
        keep_tmp: bool = False,
        link_latest: bool = True,
        promote_to_output: Iterable[str] = (
            "tree_of_thoughts.png",
            "reasoning_tree.json",
        ),
        ignore_names: Iterable[str] = (".cache", ".env"),
    ) -> Path | None:
        """Actions to perform after running the flow.

        Parameters
        ----------
        results : list[dict[str, Any]]
            The results of the flow run.
        error : BaseException | None
            Optional error during the run.
        temp_dir : Path
            The temporary directory.
        output_file : output_file : str | Path | None, optional
            The output file.
        flow_name : str
            The flow name.
        waldiez_file : Path
            The path of the waldiez file used (or dumped) for the run.
        uploads_root :  Path | None, optional
            The runtime uploads root, by default None
        skip_mmd : bool, optional
            Whether to skip the mermaid sequence diagram generation,
            by default, False
        skip_timeline : bool, optional
            Whether to skip the timeline processing, by default False
        storage_manager: StorageManager | None
            The storage manager for copying results and generating symlinks.
        metadata : dict[str, Any] | None
            Optional metadata to store in the checkpoint.
        link_latest : bool
            Whether to update a `latest` link under link_root/session_name.
            Defaults to True.
        keep_tmp : bool
            If False, delete the tmp_dir after copying. Defaults to False.
        copy_artifacts_into : str | None
            If set, copy artifacts into
            checkpoint/<copy_into_subdir> instead of the root.
        promote_to_output : Iterable[str]
            File names (exact matches) to also copy into output_dir.
        ignore_names : Iterable[str]
            Directory/file names to skip entirely.

        Returns
        -------
        Path | None
            The destination directory if output file, else None
        """
        return await anyio.to_thread.run_sync(
            ResultsMixin.post_run,
            results,
            error,
            temp_dir,
            output_file,
            flow_name,
            waldiez_file,
            uploads_root,
            skip_mmd,
            skip_timeline,
            storage_manager,
            metadata,
            copy_artifacts_into,
            keep_tmp,
            link_latest,
            promote_to_output,
            ignore_names,
        )

    @staticmethod
    def make_mermaid_diagram(
        temp_dir: Path,
        output_file: str | Path | None,
        flow_name: str,
        mmd_dir: Path,
    ) -> None:
        """Generate mermaid diagram.

        Parameters
        ----------
        temp_dir : Path
            The directory to look for logs.
        output_file : str | Path | None
            The optional destination python file.
        flow_name : str
            The name of the flow.
        mmd_dir : Path
            The path to save the mmd file to.
        """
        events_csv_path = temp_dir / "logs" / "events.csv"
        if events_csv_path.exists():
            print("Generating mermaid sequence diagram...")
            mmd_path = temp_dir / f"{flow_name}.mmd"
            generate_sequence_diagram(events_csv_path, mmd_path)
            if (
                not output_file
                and mmd_path.exists()
                and mmd_path != mmd_dir / f"{flow_name}.mmd"
            ):
                try:
                    shutil.copyfile(mmd_path, mmd_dir / f"{flow_name}.mmd")
                except BaseException:
                    pass

    @staticmethod
    def make_timeline_json(
        output_dir: Path,
    ) -> None:
        """Make the timeline JSON file.

        Parameters
        ----------
        output_dir : Path
            The path to search the events csv.
        """
        events_csv_path = output_dir / "logs" / "events.csv"
        if events_csv_path.exists():
            log_files = TimelineProcessor.get_files(output_dir / "logs")
            if any(log_files.values()):  # pragma: no branch
                output_file = output_dir / "timeline.json"
                # pylint: disable=too-many-try-statements
                try:
                    processor = TimelineProcessor()
                    processor.load_csv_files(
                        agents_file=log_files["agents"],
                        chat_file=log_files["chat"],
                        events_file=log_files["events"],
                        functions_file=log_files["functions"],
                    )
                    results = processor.process_timeline()
                    with open(
                        output_file, "w", encoding="utf-8", newline="\n"
                    ) as f:
                        json.dump(results, f, indent=2, default=str)
                    short_results = TimelineProcessor.get_short_results(results)
                    printer = get_printer()
                    printer(
                        json.dumps(
                            {"type": "timeline", "content": short_results},
                            default=str,
                        ),
                        flush=True,
                    )
                except BaseException:
                    pass

    @staticmethod
    def ensure_results_json(
        output_dir: Path,
        results: list[dict[str, Any]],
    ) -> None:
        """Ensure results.json exists in the output.

        Parameters
        ----------
        output_dir : Path
            The directory with the outputs.
        results : list[dict[str, Any]]
            The returned results.
        """
        from_json = get_results_from_json(output_dir)
        if from_json:
            store_full_results(output_dir)
            return
        remove_results_json(output_dir)
        results_json = output_dir / "results.json"
        try:
            with open(
                results_json, "w", encoding="utf-8", newline="\n"
            ) as file:
                file.write(json.dumps({"results": results}))
        except BaseException:
            return
        store_full_results(output_dir)

    @staticmethod
    async def a_ensure_results_json(
        output_dir: Path, results: list[dict[str, Any]]
    ) -> None:
        """Ensure results.json exists in the output.

        Parameters
        ----------
        output_dir : Path
            The directory with the outputs.
        results : list[dict[str, Any]]
            The returned results.
        """
        from_json = await a_get_results_from_json(output_dir)
        if from_json:
            await a_store_full_results(output_dir)
            return
        remove_results_json(output_dir)
        results_json = output_dir / "results.json"
        try:
            async with aiofiles.open(
                results_json, "w", encoding="utf-8", newline="\n"
            ) as file:
                await file.write(json.dumps({"results": results}))
        except BaseException:
            return
        await a_store_full_results(output_dir)

    @staticmethod
    def ensure_error_json(output_dir: Path, error: BaseException) -> None:
        """Ensure an error.json exists in the output.

        Parameters
        ----------
        output_dir : Path
            The path of the output
        error : BaseException
            The error that happened.
        """
        _ensure_error_json(output_dir, error)

    @staticmethod
    def get_results(
        results: list[dict[str, Any]],
        output_dir: Path,
    ) -> list[dict[str, Any]]:
        """Gather the results.

        Parameters
        ----------
        results : list[dict[str, Any]]
            The returned results from the module call.
        output_dir : Path
            The output directory to look for results.json

        Returns
        -------
        list[dict[str, Any]]
            The final results.
        """
        if (output_dir / "results.json").exists():
            return ResultsMixin.read_from_output(output_dir)
        if (output_dir / "error.json").exists():
            return ResultsMixin.read_results_error(output_dir / "error.json")
        return results

    @staticmethod
    async def a_get_results(
        results: list[dict[str, Any]],
        output_dir: Path,
    ) -> list[dict[str, Any]]:
        """Gather the results.

        Parameters
        ----------
        results : list[dict[str, Any]]
            The returned results from the module call.
        output_dir : Path
            The output directory to look for results.json

        Returns
        -------
        list[dict[str, Any]]
            The final results.
        """
        if (output_dir / "results.json").exists():
            return await ResultsMixin.a_read_from_output(output_dir)
        if (output_dir / "error.json").exists():
            return await ResultsMixin.a_read_results_error(
                output_dir / "error.json"
            )
        return results

    @staticmethod
    async def a_read_from_output(
        output_dir: Path,
    ) -> list[dict[str, Any]]:
        """Read from output dir results.json or error.json.

        Parameters
        ----------
        output_dir : Path
            The output directory to check for results.json or error.json

        Return
        ------
        list[dict[str, Any]]
            The parsed results.
        """
        error_json = output_dir / "error.json"
        results_json = output_dir / "results.json"
        try:
            if results_json.is_file():
                async with aiofiles.open(
                    results_json, "r", encoding="utf-8"
                ) as file:
                    results = await file.read()
                    return json.loads(results).get("results", [])
            if error_json.is_file():
                async with aiofiles.open(
                    error_json, "r", encoding="utf-8"
                ) as file:
                    results = await file.read()
                    reason = json.loads(results).get("error", "Flow failed")
                    return [{"error": reason}]
        except BaseException as e:
            return [{"error": str(e)}]
        return [{"error": "Could not gather result details."}]

    @staticmethod
    def read_from_output(
        output_dir: Path,
    ) -> list[dict[str, Any]]:
        """Read from output dir results.json or error.json.

        Parameters
        ----------
        output_dir : Path
            The output directory to check for results.json or error.json

        Return
        ------
        list[dict[str, Any]]
            The parsed results.
        """
        error_json = output_dir / "error.json"
        results_json = output_dir / "results.json"
        try:
            if results_json.is_file():
                with open(results_json, "r", encoding="utf-8") as file:
                    results = file.read()
                    return json.loads(results).get("results", [])
            if error_json.is_file():
                with open(error_json, "r", encoding="utf-8") as file:
                    results = file.read()
                    reason = json.loads(results).get("error", "Flow failed")
                    return [{"error": reason}]
        except BaseException as e:
            return [{"error": str(e)}]
        return [{"error": "Could not gather result details."}]

    @staticmethod
    def read_results_error(error_json: Path) -> list[dict[str, Any]]:
        """Read the error from error.json.

        Parameters
        ----------
        error_json : Path
            The path of error.json

        Returns
        -------
        list[dict[str, Any]]
            The parsed error details.
        """
        if not error_json.is_file():  # pragma: no cover
            return [{"error": "No results generated"}]
        try:
            with open(error_json, "r", encoding="utf-8") as error_file:
                error_content = error_file.read()
                error_details = json.loads(error_content)
                if isinstance(error_details, dict):
                    return [error_details]
                if isinstance(error_details, list):
                    return error_details
        except BaseException as error:
            return [{"error": str(error)}]
        return [{"error": "Failed to get error details"}]

    @staticmethod
    async def a_read_results_error(error_json: Path) -> list[dict[str, Any]]:
        """Read the error from error.json.

        Parameters
        ----------
        error_json : Path
            The path of error.json

        Returns
        -------
        list[dict[str, Any]]
            The parsed error details.
        """
        if not error_json.is_file():  # pragma: no cover
            return [{"error": "No results generated"}]
        try:
            async with aiofiles.open(
                error_json, "r", encoding="utf-8"
            ) as error_file:
                error_content = await error_file.read()
                error_details = json.loads(error_content)
                if isinstance(error_details, dict):
                    return [error_details]
                if isinstance(error_details, list):
                    return error_details
        except BaseException as error:
            return [{"error": str(error)}]
        return [{"error": "Failed to get error details"}]

    @staticmethod
    def ensure_db_outputs(output_dir: Path) -> None:
        """Ensure the csv and json files are generated if a flow.db exists.

        Parameters
        ----------
        output_dir : Path
            The output directory.
        """
        flow_db = output_dir / "flow.db"
        if not flow_db.is_file():
            return
        tables = [
            "chat_completions",
            "agents",
            "oai_wrappers",
            "oai_clients",
            "version",
            "events",
            "function_calls",
        ]
        dest = output_dir / "logs"
        dest.mkdir(parents=True, exist_ok=True)
        for table in tables:
            table_csv = dest / f"{table}.csv"
            table_json = dest / f"{table}.json"
            if not table_csv.exists() or not table_json.exists():
                get_sqlite_out(str(flow_db), table, str(table_csv))

    @staticmethod
    async def a_ensure_db_outputs(output_dir: Path) -> None:
        """Ensure the csv and json files are generated if a flow.db exists.

        Parameters
        ----------
        output_dir : Path
            The output directory.
        """
        flow_db = output_dir / "flow.db"
        if not flow_db.is_file():
            return
        tables = [
            "chat_completions",
            "agents",
            "oai_wrappers",
            "oai_clients",
            "version",
            "events",
            "function_calls",
        ]
        dest = output_dir / "logs"
        dest.mkdir(parents=True, exist_ok=True)
        for table in tables:
            table_csv = dest / f"{table}.csv"
            table_json = dest / f"{table}.json"
            if not table_csv.exists() or not table_json.exists():
                await a_get_sqlite_out(str(flow_db), table, str(table_csv))



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pyright: reportArgumentType=false
"""Reasoning agent data model."""

from typing import Any

from pydantic import Field
from typing_extensions import Annotated, Literal

from ..agent import WaldiezAgentData
from .reasoning_agent_reason_config import WaldiezReasoningAgentReasonConfig


class WaldiezReasoningAgentData(WaldiezAgentData):
    """Reasoning agent data model."""

    human_input_mode: Annotated[
        Literal["ALWAYS", "NEVER", "TERMINATE"],
        Field(
            "NEVER",
            title="Human input mode",
            description="The human input mode, Defaults to `NEVER`",
            alias="humanInputMode",
        ),
    ]
    verbose: Annotated[
        bool,
        Field(
            True,
            title="Verbose",
            description="Whether to show intermediate steps",
        ),
    ]
    reason_config: Annotated[
        WaldiezReasoningAgentReasonConfig,
        Field(
            title="Reason config",
            description="The reasoning agent's reason configuration",
            default_factory=WaldiezReasoningAgentReasonConfig,
            alias="reasonConfig",
        ),
    ]

    @property
    def method(self) -> str:
        """Get the method of the reasoning agent.

        Returns
        -------
        str
            The method of the reasoning agent.

        """
        return self.reason_config.method

    @property
    def max_depth(self) -> int:
        """Get the maximum depth of the reasoning agent.

        Returns
        -------
        int
            The maximum depth of the reasoning agent.

        """
        return self.reason_config.max_depth

    @property
    def beam_size(self) -> int:
        """Get the beam size of the reasoning agent.

        Returns
        -------
        int
            The beam size of the reasoning agent.

        """
        return self.reason_config.beam_size

    @property
    def answer_approach(self) -> str:
        """Get the answer approach of the reasoning agent.

        Returns
        -------
        str
            The answer approach of the reasoning agent.

        """
        return self.reason_config.answer_approach

    @property
    def forest_size(self) -> int:
        """Get the forest size of the reasoning agent.

        Returns
        -------
        int
            The forest size of the reasoning agent.

        """
        return self.reason_config.forest_size

    @property
    def rating_scale(self) -> int:
        """Get the rating scale of the reasoning agent.

        Returns
        -------
        int
            The rating scale of the reasoning agent.

        """
        return self.reason_config.rating_scale

    @property
    def exploration_constant(self) -> float:
        """Get the exploration constant of the reasoning agent.

        Returns
        -------
        float
            The exploration constant of the reasoning agent.

        """
        return self.reason_config.exploration_constant

    @property
    def nsim(self) -> int:
        """Get the number of simulations of the reasoning agent.

        Returns
        -------
        int
            The number of simulations of the reasoning agent.

        """
        return self.reason_config.nsim

    def get_reasoning_config(self) -> dict[str, Any]:
        """Get the reasoning configuration based on the reason_config method.

        Returns
        -------
        dict[str, Any]
            The reasoning configuration.

        """
        reason_dict: dict[str, Any] = {
            "method": self.reason_config.method,
            "max_depth": self.reason_config.max_depth,
            "forest_size": self.reason_config.forest_size,
            "rating_scale": self.reason_config.rating_scale,
        }
        if self.reason_config.method == "beam_search":
            reason_dict["beam_size"] = self.reason_config.beam_size
            reason_dict["answer_approach"] = self.reason_config.answer_approach
        if self.reason_config.method in ("mcts", "lats"):
            reason_dict["nsim"] = self.reason_config.nsim
            reason_dict["exploration_constant"] = (
                self.reason_config.exploration_constant
            )
        return reason_dict



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=line-too-long
# flake8: noqa: E501
"""Predefined YouTube search tool for Waldiez."""

from typing import Any

from ._config import PredefinedToolConfig
from .protocol import PredefinedTool


class YouTubeSearchToolImpl(PredefinedTool):
    """YouTube search tool for Waldiez."""

    required_secrets: list[str] = ["YOUTUBE_API_KEY"]
    required_kwargs: dict[str, type] = {
        "youtube_search_engine_id": str,
    }

    @property
    def name(self) -> str:
        """Tool name."""
        return "youtube_search"

    @property
    def description(self) -> str:
        """Tool description."""
        return "Search YouTube for a given query."

    @property
    def kwargs(self) -> dict[str, Any]:
        """Keyword arguments for the tool, used for initialization."""
        return {}

    @property
    def requirements(self) -> list[str]:
        """Python requirements for the tool."""
        return ["ag2[google-search]"]

    @property
    def tags(self) -> list[str]:
        """Tags for the tool, used for categorization."""
        return ["youtube", "search", "video", "reference"]

    @property
    def tool_imports(self) -> list[str]:
        """Imports required for the tool implementation."""
        return [
            "from autogen.tools.experimental import YoutubeSearchTool",
        ]

    def validate_secrets(self, secrets: dict[str, str]) -> list[str]:
        """Validate secrets and return list of missing required ones.

        Parameters
        ----------
        secrets : dict[str, str]
            Dictionary of secrets/environment variables.

        Returns
        -------
        list[str]
            List of missing required secrets.
        """
        missing: list[str] = []
        for secret in self.required_secrets:
            if secret not in secrets:
                missing.append(secret)
        return missing

    def validate_kwargs(self, kwargs: dict[str, Any]) -> list[str]:
        """Validate keyword arguments and return list of missing required ones.

        Parameters
        ----------
        kwargs : dict[str, Any]
            Dictionary of keyword arguments.

        Returns
        -------
        list[str]
            List of missing required keyword arguments.
        """
        return []

    def get_content(
        self,
        secrets: dict[str, str],
        runtime_kwargs: dict[str, Any] | None = None,
    ) -> str:
        """Get content for the tool.

        Parameters
        ----------
        secrets : dict[str, str]
            Dictionary of secrets/environment variables.
        runtime_kwargs : dict[str, Any] | None, optional
            Runtime keyword arguments to customize the content generation.

        Returns
        -------
        str
            The content for the tool.
        """
        content = f'''
def {self.name}(
    query: str,
    max_results: int = 5,
    include_video_details: bool = True,
) -> list[dict[str, Any]]:
    """Perform a YouTube search and return formatted results.

    Args:
        query: The search query string.
        max_results: The maximum number of results to return. Defaults to 5.
        include_video_details: Whether to include detailed video information. Defaults to True.

    Returns:
        A list of dictionaries of the search results.

    Raises:
        ValueError: If YOUTUBE_API_KEY is not set or if the search fails.
    """
    youtube_api_key = os.environ.get("YOUTUBE_API_KEY", "")
    if not youtube_api_key:
        raise ValueError("YOUTUBE_API_KEY is required for YouTube search tool.")
    youtube_search_tool = YoutubeSearchTool(
        youtube_api_key=youtube_api_key,
    )
    return youtube_search_tool(
        query=query,
        youtube_api_key=youtube_api_key,
        max_results=max_results,
        include_video_details=include_video_details,
    )
'''
        return content


# pylint: disable=invalid-name
YouTubeSearchTool = YouTubeSearchToolImpl()
YouTubeSearchConfig = PredefinedToolConfig(
    name=YouTubeSearchTool.name,
    description=YouTubeSearchTool.description,
    required_secrets=YouTubeSearchTool.required_secrets,
    required_kwargs=YouTubeSearchTool.required_kwargs,
    requirements=YouTubeSearchTool.requirements,
    tags=YouTubeSearchTool.tags,
    implementation=YouTubeSearchTool,
)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=too-few-public-methods,no-self-use
"""Reasoning agent configuration processor."""

from waldiez.models import WaldiezAgent, WaldiezReasoningAgent

from ...core import (
    CodeExecutionConfig,
    DefaultSerializer,
    InstanceArgument,
    Serializer,
    SystemMessageConfig,
    TerminationConfig,
)
from ...core.extras.agent_extras import ReasoningExtras


class ReasoningAgentProcessor:
    """Processor for reasoning agent configuration."""

    def __init__(self, agent: WaldiezAgent, serializer: Serializer | None):
        """Initialize the processor with the agent and serializer.

        Parameters
        ----------
        agent : WaldiezAgent
            The Waldiez agent to process.
        serializer : Serializer | None
            Optional serializer for the reasoning configuration.
            Defaults to DefaultSerializer if not provided.
        """
        self.agent = agent
        self.serializer = serializer or DefaultSerializer()

    def process(
        self,
        code_execution_config: CodeExecutionConfig | None = None,
        termination_config: TerminationConfig | None = None,
        system_message_config: SystemMessageConfig | None = None,
    ) -> ReasoningExtras:
        """Process reasoning agent configuration.

        Parameters
        ----------
        code_execution_config : CodeExecutionConfig, optional
            Configuration for code execution, if applicable.
        termination_config : TerminationConfig, optional
            Configuration for termination, if applicable.
        system_message_config : SystemMessageConfig, optional
            Configuration for system messages, if applicable.

        Returns
        -------
        ReasoningExtras
            The processed result containing extra arguments, before content,
            imports, and environment variables.
        """
        result = ReasoningExtras(
            instance_id=self.agent.id,
            code_execution_config=code_execution_config,
            termination_config=termination_config,
            system_message_config=system_message_config,
        )
        if not self.agent.is_reasoning or not isinstance(
            self.agent, WaldiezReasoningAgent
        ):  # pragma: no cover
            return result

        reasoning_config = self.agent.get_reasoning_config()
        serialized = self.serializer.serialize(reasoning_config)
        reason_arg = InstanceArgument(
            instance_id=self.agent.id,
            name="reason_config",
            value=serialized,
            tabs=1,
        )
        result.add_arg(reason_arg)
        verbose_arg = InstanceArgument(
            instance_id=self.agent.id,
            name="verbose",
            value=self.agent.data.verbose,
            tabs=1,
        )
        result.add_arg(verbose_arg)
        return result



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pyright: reportUnknownVariableType=false

"""WaldiezCheckpoint data structures."""

import json
from contextlib import suppress
from dataclasses import dataclass, field
from datetime import datetime, timezone
from pathlib import Path
from typing import Any


# noinspection PyBroadException
@dataclass
class WaldiezCheckpoint:
    """A saved checkpoint."""

    session_name: str
    timestamp: datetime
    path: Path
    _state: dict[str, Any] | None = field(init=False, default=None)
    _metadata: dict[str, Any] | None = field(init=False, default=None)

    @property
    def id(self) -> str:
        """Get the id of the checkpoint.

        Returns
        -------
        str
            The id of the checkpoint.
        """
        return WaldiezCheckpoint.format_timestamp(self.timestamp)

    @property
    def state(self) -> dict[str, Any]:
        """Get the checkpoint's state."""
        if self._state is None:
            self._state = self._load_json(self.state_file) or {}
        return self._state

    @property
    def metadata(self) -> dict[str, Any]:
        """Get the checkpoint's metadata."""
        if self._metadata is None:
            self._metadata = self._load_json(self.metadata_file) or {}
        return self._metadata

    @property
    def state_file(self) -> Path:
        """Path to the state.json file."""
        return self.path / "state.json"

    @property
    def metadata_file(self) -> Path:
        """Path to the metadata.json file."""
        return self.path / "metadata.json"

    @property
    def history_file(self) -> Path:
        """Path to the history.json file."""
        return self.path / "history.json"

    @property
    def exists(self) -> bool:
        """Check if the checkpoint exists on disk."""
        return self.path.is_dir() and self.state_file.is_file()

    def load_state(self, index: int) -> None:
        """Load a state from history from its index.

        Parameters
        ----------
        index : int
            The history index to use
        """
        entries = self.history()
        if len(entries) < index:
            return
        entry = entries[index]
        state = entry.get("state", {})
        if not isinstance(state, dict):
            return
        if "messages" in state or "context_variables" in state:
            with open(self.state_file, "w", encoding="utf-8") as f:
                json.dump(state, f)
            self.refresh()

    def history(self) -> list[dict[str, Any]]:
        """Get the state history.

        Returns
        -------
        list[dict[str, Any]]
            The stored history entries
        """
        if not self.history_file.is_file():
            return []
        history_dict = self._load_json(self.history_file)
        if not history_dict:
            return []
        history_entries = history_dict.get("history", [])
        if history_entries and isinstance(history_entries, list):
            return history_entries
        return []

    def to_dict(self, include_history: bool = False) -> dict[str, Any]:
        """Get the dict representation of the checkpoint.

        Parameters
        ----------
        include_history : bool
            Include the history in the dict or not.

        Returns
        -------
        dict[str, Any]
            The dict representation of the checkpoint.
        """
        ts = self.timestamp
        if ts.tzinfo is None:
            ts = ts.astimezone(timezone.utc)
        elif ts.tzinfo != timezone.utc:
            ts = ts.replace(tzinfo=timezone.utc)
        the_dict: dict[str, Any] = {
            "session": self.session_name,
            "timestamp": WaldiezCheckpoint.format_timestamp(ts),
            "path": self.path,
            "name": str(self.path.name),
            "state": self.state,
            "metadata": self.metadata,
        }
        if include_history:
            the_dict["history"] = self.history()
        return the_dict

    def refresh(self) -> None:
        """Reset the state and metadata."""
        self._state = None
        self._metadata = None

    @staticmethod
    def parse_timestamp(timestamp_str: str) -> datetime | None:
        """Parse timestamp from directory name.

        Parameters
        ----------
        timestamp_str : str
            The directory name

        Returns
        -------
        datetime | None
            The parsed datetime if the directory name has the expected format.
        """
        # pylint: disable=broad-exception-caught
        try:
            ts_int = int(timestamp_str)
        except Exception:
            return None
        # Normalize to seconds
        if ts_int > 1e15:  # microseconds
            ts_sec = ts_int / 1_000_000
        elif ts_int > 1e12:  # milliseconds
            ts_sec = ts_int / 1_000
        else:  # seconds
            ts_sec = ts_int
        try:
            return datetime.fromtimestamp(ts_sec, tz=timezone.utc)
        except Exception:
            return None

    @staticmethod
    def format_timestamp(timestamp: datetime) -> str:
        """Format timestamp for directory name.

        Parameters
        ----------
        timestamp: datetime
            The datetime instance.

        Returns
        -------
        str
            The formatted string.
        """
        return str(int(timestamp.timestamp() * 1_000_000))

    @staticmethod
    def _load_json(path: Path) -> dict[str, Any] | None:
        with suppress(Exception):
            with open(path, "r", encoding="utf-8") as f:
                data = json.load(f)
                if isinstance(data, dict):
                    return data
                if (
                    isinstance(data, list)
                    and data
                    and isinstance(data[0], dict)
                ):
                    return data[0]
        return None


@dataclass
class WaldiezCheckpointInfo:
    """Information about a checkpoint."""

    session_name: str
    timestamp: datetime
    path: Path
    _checkpoint: WaldiezCheckpoint | None = field(init=False, default=None)

    @property
    def id(self) -> str:
        """Get the id of the checkpoint info."""
        return WaldiezCheckpoint.format_timestamp(self.timestamp)

    @property
    def checkpoint(self) -> WaldiezCheckpoint:
        """Get the checkpoint."""
        if self._checkpoint is None:
            self._checkpoint = WaldiezCheckpoint(
                session_name=self.session_name,
                timestamp=self.timestamp,
                path=self.path,
            )
        return self._checkpoint

    def to_dict(self) -> dict[str, Any]:
        """Get the dict representation of the checkpoint's info.

        Returns
        -------
        dict[str, Any]
            The dict representation of the checkpoint's info.
        """
        ts = self.timestamp
        if ts.tzinfo is None:
            ts = ts.astimezone(timezone.utc)
        elif ts.tzinfo != timezone.utc:
            ts = ts.replace(tzinfo=timezone.utc)
        return {
            "session": self.session_name,
            "timestamp": WaldiezCheckpoint.format_timestamp(ts),
            "path": str(self.path),
            "name": str(self.path.name),
        }

    @classmethod
    def from_checkpoint(
        cls, checkpoint: WaldiezCheckpoint
    ) -> "WaldiezCheckpointInfo":
        """Create WaldiezCheckpointInfo from a WaldiezCheckpoint.

        Parameters
        ----------
        checkpoint : WaldiezCheckpoint
            The checkpoint to load info from

        Returns
        -------
        WaldiezCheckpointInfo
            The loaded checkpoint info.
        """
        return cls(
            session_name=checkpoint.session_name,
            timestamp=checkpoint.timestamp,
            path=checkpoint.path,
        )



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Models package for structured I/O."""

# Base models
from .base import PrintMessage, StructuredBase, UserInputRequest

# Constants and mappings
from .constants import (
    CONTENT_MAPPING,
    ContentMappingEntry,
    ContentTypeKey,
    MediaContent,
)

# Content models
from .content import (
    AudioContent,
    AudioMediaContent,
    FileContent,
    FileMediaContent,
    ImageContent,
    ImageMediaContent,
    ImageUrlMediaContent,
    TextMediaContent,
    VideoContent,
    VideoMediaContent,
)

# User models
from .user_input import UserInputData
from .user_response import UserResponse

__all__ = [
    # Base models
    "StructuredBase",
    "UserInputRequest",
    "PrintMessage",
    # Content base types
    "ImageContent",
    "VideoContent",
    "AudioContent",
    "FileContent",
    "MediaContent",
    # Content models
    "TextMediaContent",
    "ImageMediaContent",
    "ImageUrlMediaContent",
    "VideoMediaContent",
    "AudioMediaContent",
    "FileMediaContent",
    # User models
    "UserInputData",
    "UserResponse",
    # Constants
    "ContentTypeKey",
    "ContentMappingEntry",
    "CONTENT_MAPPING",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Utility functions for generating agent related strings."""

from .after_work import AfterWorkProcessor, AfterWorkResult
from .handoff import HandoffProcessor, HandoffResult
from .target import (
    TargetResult,
    TransitionTargetProcessor,
)

__all__ = [
    "AfterWorkProcessor",
    "AfterWorkResult",
    "HandoffProcessor",
    "HandoffResult",
    "TargetResult",
    "TransitionTargetProcessor",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pyright: reportUnknownMemberType=false,reportUnknownVariableType=false
# pyright: reportUnknownArgumentType=false, reportUnusedImport=false
"""Function related utilities."""

import ast
import importlib.util
import sys
import sysconfig
from pathlib import Path
from typing import NamedTuple

# parso for extracting function bodies
# (keeps comments, docstrings and formatting as-is)
import parso
import parso.python
import parso.tree

# let's limit the variable name length
MAX_VAR_NAME_LENGTH = 64


class ParseResult(NamedTuple):
    """Result of parsing a code string."""

    error: str | None
    tree: ast.Module | None


def is_standard_library(module_name: str) -> bool:
    """Check if the module is part of the standard library.

    Parameters
    ----------
    module_name : str
        The module name.

    Returns
    -------
    bool
        True if the module is part of the standard library.
    """
    if module_name in sys.builtin_module_names:
        return True
    try:
        spec = importlib.util.find_spec(module_name)
    except (ImportError, ValueError, ModuleNotFoundError):  # pragma: no cover
        return False
    if spec is None or not spec.origin:  # pragma: no cover
        return False
    if "site-packages" in spec.origin:
        return False
    if spec.origin == "frozen":
        return True
    stdlib_path = Path(sysconfig.get_path("stdlib")).resolve()
    return Path(spec.origin).resolve().is_relative_to(stdlib_path)


def parse_code_string(
    code_string: str,
) -> ParseResult:
    """Parse the code string.

    Parameters
    ----------
    code_string : str
        The code string.

    Returns
    -------
    ParseResult
        If valid, None and the ast module.
        If invalid, the error message and None.
    """
    # pylint: disable=broad-exception-caught
    try:
        tree = ast.parse(code_string)
    except SyntaxError as e:
        return ParseResult(
            f"SyntaxError: {e}, in " + "\n" + f"{code_string}", None
        )
    except BaseException as e:  # pragma: no cover
        return ParseResult(
            f"Invalid code: {e}, in " + "\n" + f"{code_string}", None
        )
    return ParseResult(None, tree)


def _extract_module_name(node: ast.AST) -> str | None:
    """Extract the root module name from an import node."""
    if isinstance(node, ast.Import):
        return node.names[0].name.split(".")[0]
    if isinstance(node, ast.ImportFrom):
        if node.module:
            # Handle relative imports
            if node.module.startswith("."):  # pragma: no cover
                return None  # Skip relative imports for stdlib check
            return node.module.split(".")[0]
    return None  # pragma: no cover


def _extract_imports_from_ast(code_string: str) -> tuple[list[str], list[str]]:
    """Extract import statements from code using AST.

    Parameters
    ----------
    code_string : str
        The code string to parse.

    Returns
    -------
    tuple[list[str], list[str]]
        Standard library imports and third party imports (unsorted).
    """
    standard_lib_imports: list[str] = []
    third_party_imports: list[str] = []

    try:
        tree = ast.parse(code_string)
    except SyntaxError:  # pragma: no cover
        return [], []

    for node in ast.walk(tree):
        if isinstance(node, (ast.Import, ast.ImportFrom)):
            # noinspection PyTypeChecker
            full_import_statement = ast.get_source_segment(code_string, node)
            if not full_import_statement:  # pragma: no cover
                continue
            full_import_statement = full_import_statement.strip()

            # noinspection PyTypeChecker
            module_name = _extract_module_name(node)
            if not module_name:  # pragma: no cover
                continue

            if is_standard_library(module_name):
                standard_lib_imports.append(full_import_statement)
            else:
                third_party_imports.append(full_import_statement)

    return standard_lib_imports, third_party_imports


def _sort_imports(imports: list[str]) -> list[str]:
    """Sort import statements with 'import' statements before 'from' statements.

    Parameters
    ----------
    imports : list[str]
        List of import statements to sort.

    Returns
    -------
    list[str]
        Sorted import statements.
    """
    import_statements = sorted(
        [stmt for stmt in imports if stmt.startswith("import ")]
    )
    from_statements = sorted(
        [stmt for stmt in imports if stmt.startswith("from ")]
    )
    return import_statements + from_statements


def gather_code_imports(
    code_string: str,
    is_interop: bool,
) -> tuple[list[str], list[str]]:
    """Gather the imports from the code string.

    Parameters
    ----------
    code_string : str
        The code string.
    is_interop : bool
        If True, make sure the interoperability import is present.

    Returns
    -------
    tuple[list[str], list[str]]
        The standard library imports and the third party imports.
    """
    standard_lib_imports, third_party_imports = _extract_imports_from_ast(
        code_string
    )

    if is_interop and (
        "from autogen.interop import Interoperability"
        not in third_party_imports
    ):
        third_party_imports.append(
            "from autogen.interop import Interoperability"
        )

    sorted_standard_lib_imports = _sort_imports(standard_lib_imports)
    sorted_third_party_imports = _sort_imports(third_party_imports)

    return sorted_standard_lib_imports, sorted_third_party_imports


def check_function(
    code_string: str,
    function_name: str,
    function_args: list[str],
) -> tuple[bool, str]:
    """Check the function.

    Parameters
    ----------
    code_string : str
        The code string to check.
    function_name : str
        The expected method name.
    function_args : list[str]
        The expected method arguments.

    Returns
    -------
    tuple[bool, str]
        If valid, True and the function body (only), no extra lines.
        If invalid, False and the error message.
    """
    error, tree = parse_code_string(code_string)
    if error is not None or tree is None:
        return False, error or "Invalid code"
    return _validate_function_body(
        tree,
        code_string,
        function_name,
        function_args,
    )


def _validate_function_body(
    tree: ast.Module,
    code_string: str,
    function_name: str,
    function_args: list[str],
) -> tuple[bool, str]:
    """Get the function body.

    Parameters
    ----------
    tree : ast.Module
        The ast module.
    code_string : str
        The function body.
    function_name : str
        The expected method name.
    function_args : list[str]
        The expected method arguments.

    Returns
    -------
    tuple[bool, str]
        If valid, True and the function body (only), no extra lines.
        If invalid, False and the error message.
    """
    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef):
            if node.name != function_name:
                continue
            if len(node.args.args) != len(function_args):
                return (
                    False,
                    (
                        f"Invalid number of arguments, in function {node.name},"
                        f" expected: {len(function_args)},"
                        f" got: {len(node.args.args)} :("
                    ),
                )
            for arg, expected_arg in zip(
                node.args.args,
                function_args,
                strict=False,
            ):
                if arg.arg != expected_arg:
                    return (
                        False,
                        (
                            f"Invalid argument name: {arg.arg}"
                            f" in function {node.name}"
                        ),
                    )
            if not node.body:  # pragma: no cover
                return False, "No body found in the function"
            function_body = _get_function_body(code_string, node)
            return True, function_body
    error_msg = (
        f"No method with name `{function_name}`"
        f" and arguments `{function_args}` found"
    )
    return False, error_msg


def get_function(
    code_string: str,
    function_name: str,
) -> str:
    """Get the function signature and body.

    Parameters
    ----------
    code_string : str
        The code string.
    function_name : str
        The function name.

    Returns
    -------
    str
        The function signature and body.
    """
    tree = parso.parse(code_string)  # type: ignore
    for node in tree.iter_funcdefs():
        if node.name.value == function_name:
            return node.get_code()
    return ""


def _get_function_body(
    code_string: str,
    node: ast.FunctionDef,
) -> str:
    """Get the function body, including docstring and comments inside.

    Parameters
    ----------
    code_string : str
        The code string.
    node : ast.FunctionDef
        The function node.

    Returns
    -------
    str
        The function body.

    Raises
    ------
    ValueError
        If the function's body is empty.
    """
    lines = code_string.splitlines()
    signature_start_line = node.lineno - 1
    body_start_line = node.body[0].lineno - 1
    signature_end_line = signature_start_line
    for i in range(signature_start_line, body_start_line):  # pragma: no branch
        if ")" in lines[i]:
            signature_end_line = i
            break
    function_body_lines = lines[signature_end_line + 1 :]
    last_line = function_body_lines[-1]
    if not last_line.strip() and len(function_body_lines) > 1:
        function_body_lines = function_body_lines[:-1]
    function_body = "\n".join(function_body_lines)
    while function_body.startswith("\n"):
        function_body = function_body[1:]
    while function_body.endswith("\n"):
        function_body = function_body[:-1]
    return function_body


def generate_function(
    function_name: str,
    function_args: list[str],
    function_types: tuple[list[str], str],
    function_body: str,
    types_as_comments: bool = False,
) -> str:
    """Generate a function.

    Parameters
    ----------
    function_name : str
        The function name.
    function_args : list[str]
        The function arguments.
    function_types : tuple[list[str], str]
        The function types.
    function_body : str
        The function body.
    types_as_comments : bool, optional
        Include the type hints as comments (or in the function signature)
        (default is False).

    Returns
    -------
    str
        The generated function.
    """
    if len(function_name) > MAX_VAR_NAME_LENGTH:
        function_name = function_name[:MAX_VAR_NAME_LENGTH]
    function_string = f"def {function_name}("
    if not function_args:
        function_string += ")"
    else:
        function_string += "\n"
        for arg, arg_type in zip(
            function_args, function_types[0], strict=False
        ):
            if types_as_comments:
                function_string += f"    {arg},  # type: {arg_type}" + "\n"
            else:
                function_string += f"    {arg}: {arg_type}," + "\n"
        function_string += ")"
    if types_as_comments:
        function_string += ":\n"
        function_string += "    # type: (...) -> " + function_types[1]
    else:
        function_string += " -> " + function_types[1] + ":"
    function_string += "\n" if not function_body.startswith("\n") else ""
    function_string += f"{function_body}"
    if not function_string.endswith("\n"):  # pragma: no branch
        function_string += "\n"
    return function_string



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Agent exporter rag extras."""

from .vector_db_extras import VectorDBExtras, get_vector_db_extras

__all__ = [
    "get_vector_db_extras",
    "VectorDBExtras",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Ensure unique names for agents, models, tools, and chats."""

import re

MAX_VARIABLE_LENGTH = 46


def get_valid_python_variable_name(
    possible: str,
    prefix: str = "w",
    max_length: int = MAX_VARIABLE_LENGTH,
) -> str:
    """Get a valid Python variable name from a possible name.

    Parameters
    ----------
    possible : str
        The possible name.
    prefix : str, optional
        The prefix to use if the name starts with a digit or special character
    max_length : int, optional
        The maximum length of the variable name.

    Returns
    -------
    str
        The valid Python variable name.
    """
    if not possible or not possible.strip():
        return prefix + "_"

    # First handle arrow operators specifically
    possible = possible.replace("->", "to")
    possible = possible.replace("=>", "to")
    possible = possible.replace("<-", "from")
    possible = possible.replace("<=", "from")

    # Replace non-ASCII characters and non-word characters with underscores
    # \W matches any non-word character, but in Python's re module,
    # \w includes Unicode letters by default, so we need to be more explicit
    # to replace Unicode letters with underscores for valid Python identifiers
    # noinspection RegExpSimplifiable
    possible = re.sub(r"[^\w]", "_", possible)  # Replace non-word chars
    possible = re.sub(r"[^\x00-\x7F]", "_", possible)  # Replace non-ASCII chars

    # Convert to lowercase and truncate
    possible = possible.lower()[:max_length]

    # Remove trailing underscores from truncation
    possible = possible.rstrip("_")

    if not possible:
        return prefix + "_"

    # Handle names starting with underscore
    if possible.startswith("_"):
        return f"{prefix}{possible}"

    # Handle names starting with digit
    if possible[0].isdigit():
        return f"{prefix}_{possible}"

    return possible


def get_valid_instance_name(
    instance: tuple[str, str],
    current_names: dict[str, str],
    prefix: str = "w",
    max_length: int = MAX_VARIABLE_LENGTH,
) -> dict[str, str]:
    """Get a valid instance name.

    If the instance id is already in the current names nothing is done.
    If the name already exists in the current names,
        the name is updated (with an index suffix).

    Parameters
    ----------
    instance : tuple[str, str]
        The instance id and possible name.
    current_names : dict[str, str]
        The current names.
    prefix : str, optional
        The prefix to use if the name starts with a digit,
        if the name is already in the current names,
        or if the name is already in the current names with an index suffix.
    max_length : int, optional
        The maximum length of the variable name.

    Returns
    -------
    dict[str, str]
        The updated names.
    """
    instance_id, possible_name = instance[0], instance[1][:max_length]

    if instance_id in current_names:
        # already in the current names (by its id)
        return current_names

    new_names = current_names.copy()
    name = get_valid_python_variable_name(
        possible_name, prefix=prefix, max_length=max_length
    )

    if name in current_names.values():
        name = f"{prefix}_{name}"

    if name in current_names.values():
        index = 1
        while f"{name}_{index}" in current_names.values():
            index += 1
        name = f"{name}_{index}"

    new_names[instance_id] = name
    return new_names


def safe_filename(name: str, ext: str = "") -> str:
    """
    Make a safe cross-platform filename from an arbitrary string.

    Parameters
    ----------
    name : str
        The string to turn into a safe filename.
    ext :  str
        Optional extension (with or without leading dot).

    Returns
    -------
    str
        A safe filename string.
    """
    # Normalize extension
    # pylint: disable=inconsistent-quotes
    ext = f".{ext.lstrip('.')}" if ext else ""

    # Forbidden characters on Windows (also bad on Unix)
    forbidden = r'[<>:"/\\|?*\x00-\x1F]'
    name = re.sub(forbidden, "_", name)

    # Trim trailing dots/spaces (illegal on Windows)
    name = name.rstrip(". ")

    # Collapse multiple underscores
    name = re.sub(r"_+", "_", name)

    # Reserved Windows device names
    reserved = re.compile(
        r"^(con|prn|aux|nul|com[1-9]|lpt[1-9])$", re.IGNORECASE
    )
    if reserved.match(name):
        name = f"_{name}"

    # Fallback if empty
    if not name:
        name = "file"

    # Ensure length limit (NTFS max filename length = 255 bytes)
    max_len = 255 - len(ext)
    if len(name) > max_len:
        name = name[:max_len]

    return f"{name}{ext}"



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Flow exporter utils."""

from .common import (
    generate_header,
    get_after_run_content,
    get_common_env_var_setup,
    get_np_no_nep50_handle,
)
from .importing import get_sorted_imports, get_the_imports_string
from .logging import get_sqlite_out, get_start_logging, get_stop_logging

__all__ = [
    "generate_header",
    "get_common_env_var_setup",
    "get_np_no_nep50_handle",
    "get_after_run_content",
    "get_the_imports_string",
    "get_sorted_imports",
    "get_start_logging",
    "get_stop_logging",
    "get_sqlite_out",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez Agent types."""

from typing_extensions import Literal

# pylint: disable=line-too-long
# fmt: off
WaldiezAgentType = Literal["user_proxy", "assistant", "group_manager", "manager", "rag_user", "swarm", "reasoning", "captain", "user", "rag_user_proxy", "doc_agent"]  # noqa: E501
"""Possible types of a Waldiez Agent:
    - user_proxy,
    - assistant,
    - group_manager,
    - rag_user_proxy (deprecated: use doc_agent),
    - reasoning,
    - captain,
    - doc_agent,
    - swarm (deprecated: do not use it),
    - user (deprecated: use user_proxy)
    - rag_user (deprecated: user rag_user_proxy)
    - manager (deprecated: use group_manager)
"""  # noqa: W291
# fmt: on
